# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from collections.abc import Iterable
from dataclasses import dataclass, field
from enum import Enum, auto
from itertools import chain
from types import NotImplementedType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Self,
    Sequence,
    TypeGuard,
    cast,
    override,
)

from faebryk.core.core import Namespace
from faebryk.core.graphinterface import GraphInterface
from faebryk.core.node import Node, f_field
from faebryk.core.trait import Trait
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval,
    Quantity_Interval_Disjoint,
    Quantity_Set,
    Quantity_Set_Discrete,
    QuantityLikeR,
)
from faebryk.libs.sets.sets import BoolSet, EnumSet, P_Set, as_lit
from faebryk.libs.units import (
    HasUnit,
    Quantity,
    Unit,
    UnitCompatibilityError,
    assert_compatible_units,
    dimensionless,
    quantity,
)
from faebryk.libs.util import (
    KeyErrorAmbiguous,
    KeyErrorNotFound,
    abstract,
    cast_assert,
    find,
    once,
    unique,
)

if TYPE_CHECKING:
    from faebryk.core.solver.solver import Solver

logger = logging.getLogger(__name__)


class ParameterOperableException(Exception):
    def __init__(self, parameter: "ParameterOperatable", msg: str):
        self.parameter = parameter
        super().__init__(msg)


class ParameterOperableHasNoLiteral(ParameterOperableException):
    pass


# When we make this generic, two types, type T of elements, and type S of known subsets
# boolean: T == S == bool
# enum: T == S == Enum
# number: T == Number type, S == Range[Number]
class ParameterOperatable(Node):
    type QuantityLike = Quantity | Unit | NotImplementedType
    type Number = int | float | QuantityLike

    type NumberLiteral = Number | P_Set[Number] | Quantity_Set
    type NumberLike = ParameterOperatable | NumberLiteral
    type BooleanLiteral = bool | BoolSet
    type BooleanLike = ParameterOperatable | BooleanLiteral
    type EnumLiteral = Enum | EnumSet
    type EnumLike = ParameterOperatable | EnumLiteral
    type SetLiteral = NumberLiteral | BooleanLiteral | EnumLiteral

    type All = NumberLike | BooleanLike | EnumLike
    type Literal = NumberLiteral | BooleanLiteral | EnumLiteral | SetLiteral
    type Sets = All

    operated_on: GraphInterface

    constrained: bool = False
    """
    Shortcut for `isinstance(self, ConstrainableExpression) and self.constrained`
    """

    @property
    def domain(self) -> "Domain": ...

    def has_implicit_constraint(self) -> bool: ...

    def has_implicit_constraints_recursive(self) -> bool: ...

    def operation_add(self, other: NumberLike):
        return Add(self, other)

    def operation_subtract(self: NumberLike, other: NumberLike):
        return Subtract(self, other)

    def operation_multiply(self, other: NumberLike):
        return Multiply(self, other)

    def operation_divide(self: NumberLike, other: NumberLike):
        return Divide(self, other)

    def operation_power(self, other: NumberLike):
        return Power(base=self, exponent=other)

    def operation_log(self):
        return Log(self)

    def operation_sqrt(self):
        return Sqrt(self)

    def operation_abs(self):
        return Abs(self)

    def operation_min(self):
        return Min(self)

    def operation_max(self):
        return Max(self)

    def operation_integrate(self, variable: "Parameter"):
        return Integrate(self, variable)

    def operation_differentiate(self, variable: "Parameter"):
        return Differentiate(self, variable)

    def operation_negate(self):
        return Multiply(self, quantity(-1))

    def operation_invert(self):
        return Power(self, -1)

    def operation_floor(self):
        return Floor(self)

    def operation_ceil(self):
        return Ceil(self)

    def operation_round(self):
        return Round(self)

    def operation_sin(self):
        return Sin(self)

    def operation_cos(self):
        return Cos(self)

    def operation_union(self, other: Sets):
        return Union(self, other)

    def operation_intersection(self, other: Sets):
        return Intersection(self, other)

    def operation_difference(self, *subtrahends: Sets):
        return Difference(minuend=self, *subtrahends)

    def operation_symmetric_difference(self, other: Sets):
        return SymmetricDifference(self, other)

    def operation_and(self, other: BooleanLike):
        return And(self, other)

    def operation_or(self, other: BooleanLike):
        return Or(self, other)

    def operation_not(self):
        return Not(self)

    def operation_xor(self, other: BooleanLike):
        return Xor(self, other)

    def operation_implies(self, other: BooleanLike):
        return Implies(condition=self, implication=other)

    def operation_is_le(self, other: NumberLike):
        return LessOrEqual(left=self, right=other)

    def operation_is_ge(self, other: NumberLike):
        return GreaterOrEqual(left=self, right=other)

    def operation_is_lt(self, other: NumberLike):
        return LessThan(left=self, right=other)

    def operation_is_gt(self, other: NumberLike):
        return GreaterThan(left=self, right=other)

    def operation_is_ne(self, other: NumberLike):
        return NotEqual(left=self, right=other)

    def operation_is_subset(self, other: Sets):
        return IsSubset(left=self, right=other)

    def operation_is_superset(self, other: Sets):
        return IsSuperset(left=self, right=other)

    # Run by the solver on finalization
    inspect_solution: Callable[[Self], None] = lambda _: None

    def inspect_add_on_solution(self, fun: Callable[[Self], None]) -> None:
        current = self.inspect_solution

        def new(self2):
            current(self2)
            fun(self2)

        self.inspect_solution = new

    # ----------------------------------------------------------------------------------
    # Generic
    def alias_is(self, other: All):
        return Is(self, other).constrain()

    # Numberlike
    def constrain_le(self, other: NumberLike):
        return self.operation_is_le(other).constrain()

    def constrain_ge(self, other: NumberLike):
        return self.operation_is_ge(other).constrain()

    def constrain_lt(self, other: NumberLike):
        # makes implementation easier for now
        # le should be enough
        raise NotImplementedError()
        return self.operation_is_lt(other).constrain()

    def constrain_gt(self, other: NumberLike):
        # makes implementation easier for now
        # ge should be enough
        raise NotImplementedError()
        return self.operation_is_gt(other).constrain()

    def constrain_ne(self, other: NumberLike):
        # want to see when this is useful in practice
        raise NotImplementedError()
        return self.operation_is_ne(other).constrain()

    # Setlike
    def constrain_subset(self, other: Sets):
        return self.operation_is_subset(other).constrain()

    def constrain_superset(self, other: Sets):
        return self.operation_is_superset(other).constrain()

    def constrain_cardinality(self, other: int):
        return Cardinality(self, other).constrain()

    # ----------------------------------------------------------------------------------
    def __add__(self, other: NumberLike):
        return self.operation_add(other)

    def __radd__(self, other: NumberLike):
        return self.operation_add(other)

    def __sub__(self, other: NumberLike):
        # TODO could be set difference
        return self.operation_subtract(other)

    def __rsub__(self, other: NumberLike):
        return type(self).operation_subtract(other, self)

    def __mul__(self, other: NumberLike):
        return self.operation_multiply(other)

    def __rmul__(self, other: NumberLike):
        return self.operation_multiply(other)

    def __truediv__(self, other: NumberLike):
        return self.operation_divide(other)

    def __rtruediv__(self, other: NumberLike):
        return type(self).operation_divide(other, self)

    def __pow__(self, other: NumberLike):
        return self.operation_power(other)

    def __abs__(self):
        return self.operation_abs()

    def __neg__(self):
        return self.operation_negate()

    def __round__(self):
        return self.operation_round()

    def __floor__(self):
        return self.operation_floor()

    def __ceil__(self):
        return self.operation_ceil()

    def __le__(self, other: NumberLike):
        return self.operation_is_le(other)

    def __ge__(self, other: NumberLike):
        return self.operation_is_ge(other)

    def __lt__(self, other: NumberLike):
        return self.operation_is_lt(other)

    def __gt__(self, other: NumberLike):
        return self.operation_is_gt(other)

    def __ne__(self, other: NumberLike):
        return self.operation_is_ne(other)

    # bitwise and
    def __and__(self, other: BooleanLike):
        # TODO could be set intersection
        return self.operation_and(other)

    def __rand__(self, other: BooleanLike):
        return self.operation_and(other)

    def __or__(self, other: BooleanLike):
        # TODO could be set union
        return self.operation_or(other)

    def __ror__(self, other: BooleanLike):
        return self.operation_or(other)

    def __xor__(self, other: BooleanLike):
        return self.operation_xor(other)

    def __rxor__(self, other: BooleanLike):
        return self.operation_xor(other)

    # ----------------------------------------------------------------------------------

    def operation_switch_case_subset(
        self,
        cases: Iterable[tuple[Literal, "ConstrainableExpression"]],
    ) -> "ConstrainableExpression":
        exprs = [(IsSubset(self, lit), case) for lit, case in cases]
        return ConstrainableExpression.operation_switch_case_implications(exprs)

    def operation_mapping(
        self,
        other: "ParameterOperatable",
        mapping: dict[Literal, Literal],
    ) -> "ConstrainableExpression":
        exprs = [
            (lit_self, IsSubset(other, lit_other))
            for lit_self, lit_other in mapping.items()
        ]
        return self.operation_switch_case_subset(exprs)

    def constrain_mapping(
        self, other: "ParameterOperatable", mapping: dict[Literal, Literal]
    ):
        return self.operation_mapping(other, mapping).constrain()

    # ----------------------------------------------------------------------------------
    def get_operations[T: "Expression"](
        self,
        types: type[T] | None = None,
        constrained_only: bool = False,
        recursive: bool = False,
    ) -> set[T]:
        if types is None:
            types = Expression  # type: ignore
        types = cast(type[T], types)
        assert issubclass(types, Expression)

        out = cast(set[T], self.operated_on.get_connected_nodes(types=[types]))
        if recursive:
            for o in list(out):
                out.update(o.get_operations(recursive=True))
        if constrained_only:
            assert issubclass(types, ConstrainableExpression)
            out = {i for i in out if cast(ConstrainableExpression, i).constrained}
        return out

    def get_literal(self, op: type["ConstrainableExpression"] | None = None) -> Literal:
        """
        ```
        P(self, X) ∧ P.constrained -> return X
        ```
        """
        if op is None:
            op = Is
        # we have logic for is and is_subset in this function,
        # let's find a usecase for other ops before allowing them
        if not issubclass(op, (Is, IsSubset)):
            raise ValueError(f"Unsupported op {op}")
        ops = self.get_operations(op, constrained_only=True)
        try:
            if op is IsSubset:
                literals = find(
                    lit
                    for op in ops
                    for (i, lit) in op.get_operand_literals().items()
                    if i > 0
                )
            else:
                literals = find(
                    lit for op in ops for (_, lit) in op.get_operand_literals().items()
                )
        except KeyErrorNotFound as e:
            raise ParameterOperableHasNoLiteral(
                self, f"Parameter {self} has no literal for op {op}"
            ) from e
        except KeyErrorAmbiguous as e:
            duplicates = e.duplicates
            if issubclass(op, Is):
                if len(unique(duplicates, key=lambda x: as_lit(x))) != 1:
                    raise
                return duplicates[0]
            elif issubclass(op, IsSubset):
                return P_Set.intersect_all(*duplicates)
            else:
                raise
        return literals

    def try_get_literal_subset(self) -> Literal | None:
        lits = self.try_get_literal_for_multiple_ops([Is, IsSubset])
        if not lits:
            return None
        if len(lits) == 1:
            return next(iter(lits.values()))

        is_lit, ss_lit = map(P_Set.from_value, (lits[Is], lits[IsSubset]))
        if not is_lit.is_subset_of(ss_lit):
            raise KeyErrorAmbiguous(list(lits.values()))
        return is_lit

    def try_get_literal(
        self, op: type["ConstrainableExpression"] | None = None
    ) -> Literal | None:
        try:
            return self.get_literal(op)
        except ParameterOperableHasNoLiteral:
            return None

    def try_get_literal_for_multiple_ops(
        self, ops: list[type["ConstrainableExpression"]]
    ) -> dict[type["ConstrainableExpression"], Literal] | None:
        lits = {op: self.try_get_literal(op) for op in ops}
        lits = {op: lit for op, lit in lits.items() if lit is not None}
        return lits if lits else None

    @staticmethod
    def try_extract_literal(
        po: "ParameterOperatable.All", allow_subset: bool = False
    ) -> Literal | None:
        if ParameterOperatable.is_literal(po):
            return po
        assert isinstance(po, ParameterOperatable)
        if allow_subset:
            return po.try_get_literal_subset()
        return po.try_get_literal()

    # type checks

    # TODO Quantity_Interval_Disjoint is also a literal
    @staticmethod
    def is_number_literal(value: Any) -> TypeGuard[QuantityLike]:
        return isinstance(value, QuantityLikeR)

    @staticmethod
    def is_literal(value: Any) -> TypeGuard[Literal]:
        return not isinstance(value, ParameterOperatable)

    @dataclass
    class ReprContext:
        @dataclass
        class VariableMapping:
            mapping: dict["Parameter", int] = field(default_factory=dict)
            next_id: int = 0

        variable_mapping: VariableMapping = field(default_factory=VariableMapping)

        def __hash__(self) -> int:
            return hash(id(self))

    @once
    def compact_repr(
        self, context: ReprContext | None = None, use_name: bool = False
    ) -> str:
        raise NotImplementedError()

    # TODO move to Expression
    @staticmethod
    def sort_by_depth[T: ParameterOperatable](
        exprs: Iterable[T], ascending: bool
    ) -> list[T]:
        """
        Ascending:
        ```
        (A + B) + (C + D)
        -> [(A+B), (C+D), (A+B)+(C+D)]
        ```
        """
        return sorted(exprs, key=ParameterOperatable.get_depth, reverse=not ascending)

    @staticmethod
    def get_depth(po: "ParameterOperatable.All") -> int:
        if isinstance(po, Expression):
            return po.depth()
        return 0

    @once
    def _get_lit_suffix(self) -> str:
        out = ""
        try:
            lit = self.try_get_literal()
        except KeyErrorAmbiguous as e:
            return f"{{AMBIGUOUS_I: {e.duplicates}}}"
        if lit is not None:
            out = f"{{I|{lit}}}"
        elif (lit := self.try_get_literal_subset()) is not None:
            out = f"{{S|{lit}}}"
        if lit == BoolSet(True):
            out = "✓"
        elif lit == BoolSet(False) and isinstance(lit, (BoolSet, bool)):
            out = "✗"
        return out


def has_implicit_constraints_recursive(po: ParameterOperatable.All) -> bool:
    if isinstance(po, ParameterOperatable):
        return po.has_implicit_constraints_recursive()
    return False


@abstract
class Expression(ParameterOperatable):
    operates_on: GraphInterface

    def __init__(self, domain, *operands: ParameterOperatable.All):
        super().__init__()
        assert not any(o is None for o in operands)
        self._domain = domain
        self.operands = tuple(operands)
        self.operatable_operands: set[ParameterOperatable] = {
            op for op in operands if isinstance(op, ParameterOperatable)
        }
        self.non_operands: list[Any] = []

    def __preinit__(self):
        for op in self.operatable_operands:
            # TODO: careful here, we should make it more clear that operates_on just
            # expresses a relation but is not a replacement of self.operands
            if self.operates_on.is_connected_to(op.operated_on):
                continue
            self.operates_on.connect(op.operated_on)

    @staticmethod
    def is_uncorrelatable_literal(lit) -> bool:
        return not isinstance(lit, ParameterOperatable) and (
            not isinstance(lit, P_Set)
            or not (lit.is_single_element() or lit.is_empty())
        )

    @once
    def get_uncorrelatable_literals(self) -> list[ParameterOperatable.Literal]:
        # TODO we should just use the canonical lits, for now just no support
        # for non-canonical lits
        return [
            lit for lit in self.operands if Expression.is_uncorrelatable_literal(lit)
        ]

    @once
    def get_sorted_operands(self) -> list[ParameterOperatable]:
        return sorted(self.operands, key=hash)

    @once
    def is_congruent_to(
        self,
        other: "Expression",
        recursive: bool = False,
        allow_uncorrelated: bool = False,
        check_constrained: bool = True,
    ) -> bool:
        if self is other:
            return True
        if type(self) is not type(other):
            return False
        if len(self.operands) != len(other.operands):
            return False
        # if lit is non-single/empty set we can't correlate thus can't be congruent
        #  in general
        if not allow_uncorrelated and (
            self.get_uncorrelatable_literals() or other.get_uncorrelatable_literals()
        ):
            return False
        if (
            check_constrained
            and isinstance(self, ConstrainableExpression)
            and self.constrained
            != cast_assert(ConstrainableExpression, other).constrained
        ):
            return False
        if self.operands == other.operands:
            return True
        if isinstance(self, Commutative):
            # fucking genius
            # lit hash is stable
            # paramop hash only same with same id
            left = self.get_sorted_operands()
            right = other.get_sorted_operands()
            if left == right:
                return True

        if recursive and Expression.are_pos_congruent(
            self.operands,
            other.operands,
            commutative=isinstance(self, Commutative),
            allow_uncorrelated=allow_uncorrelated,
            check_constrained=check_constrained,
        ):
            return True

        return False

    def is_congruent_to_factory(
        self,
        other_factory: type["Expression"],
        other_operands: Sequence[ParameterOperatable.All],
        allow_uncorrelated: bool = False,
        # TODO
        check_constrained: bool = True,
    ) -> bool:
        if type(self) is not other_factory:
            return False
        return Expression.are_pos_congruent(
            self.operands,
            other_operands,
            allow_uncorrelated=allow_uncorrelated,
            check_constrained=check_constrained,
        )

    @staticmethod
    def are_pos_congruent(
        left: Sequence[ParameterOperatable.All],
        right: Sequence[ParameterOperatable.All],
        commutative: bool = False,
        allow_uncorrelated: bool = False,
        check_constrained: bool = True,
    ) -> bool:
        if len(left) != len(right):
            return False

        if not allow_uncorrelated and any(
            Expression.is_uncorrelatable_literal(lit) for lit in chain(left, right)
        ):
            return False

        # FIXME handle commutative
        return all(
            lhs.is_congruent_to(
                rhs,
                recursive=True,
                allow_uncorrelated=allow_uncorrelated,
                check_constrained=check_constrained,
            )
            if isinstance(lhs, Expression) and isinstance(rhs, Expression)
            else lhs == rhs
            for lhs, rhs in zip(left, right)
        )

    @property
    def domain(self) -> "Domain":
        return self._domain

    def get_operand_operatables[T: ParameterOperatable](
        self, types: type[T] = ParameterOperatable, recursive: bool = False
    ) -> set[T]:
        out = set(self.operatable_operands)
        if recursive:
            for o in list(out):
                if not isinstance(o, Expression):
                    continue
                out.update(o.get_operand_operatables(recursive=True))
        if types is not ParameterOperatable:
            out = {o for o in out if isinstance(o, types)}
        return cast(set[T], out)

    def get_operand_expressions(self, recursive: bool = False) -> set["Expression"]:
        return self.get_operand_operatables(Expression, recursive=recursive)

    def get_operand_parameters(self, recursive: bool = False) -> set["Parameter"]:
        return self.get_operand_operatables(Parameter, recursive=recursive)

    def get_operand_literals(self) -> dict[int, ParameterOperatable.Literal]:
        return {
            i: o
            for i, o in enumerate(self.operands)
            if ParameterOperatable.is_literal(o)
        }

    def get_operand_leaves_operatable(self) -> set[ParameterOperatable]:
        ops = self.get_operand_operatables(recursive=True)
        return {
            o
            for o in ops
            if not isinstance(o, Expression)
            or not o.get_operand_operatables(recursive=False)
        }

    def depth(self) -> int:
        """
        Returns depth of longest expression tree from this expression.
        ```
        ((A + B) + (C + D)) * 5
            ^    ^    ^     ^
            0    1    0     2

        a = (X + (Y + Z))
        (a + 1) + a
         ^ ^    ^ ^
         1 2    3 1
        ```
        """
        # FIXME this does not work (if expressions are added afterwards in the tree)
        if hasattr(self, "_depth"):
            return self._depth
        self._depth = 1 + max(
            [0] + [Expression.get_depth(op) for op in self.operands],
        )
        return self._depth

    def has_implicit_constraint(self) -> bool:
        return True  # opt out

    def has_implicit_constraints_recursive(self) -> bool:
        if self.has_implicit_constraint():
            return True
        for op in self.operands:
            if isinstance(op, Expression) and op.has_implicit_constraints_recursive():
                return True
        return False

    def get_other_operand(
        self, operand: ParameterOperatable.All
    ) -> ParameterOperatable.All:
        assert len(self.operands) == 2
        if self.operands[0] is operand:
            return self.operands[1]
        return self.operands[0]

    def __repr__(self) -> str:
        return f"{super().__repr__()}({self.operands})"

    @dataclass
    class ReprStyle:
        symbol: str | None = None

        class Placement(Enum):
            INFIX = auto()
            """
            A + B + C
            """
            INFIX_FIRST = auto()
            """
            A > (B, C)
            """
            PREFIX = auto()
            """
            ¬A
            """
            POSTFIX = auto()
            """
            A!
            """
            EMBRACE = auto()
            """
            |A|
            """

        placement: Placement = Placement.INFIX

    REPR_STYLE: ReprStyle = ReprStyle()

    def compact_repr(
        self,
        context: ParameterOperatable.ReprContext | None = None,
        use_name: bool = False,
    ) -> str:
        if context is None:
            context = ParameterOperatable.ReprContext()

        style = type(self).REPR_STYLE
        symbol = style.symbol
        if symbol is None:
            symbol = type(self).__name__

        symbol_suffix = ""
        if isinstance(self, ConstrainableExpression) and self.constrained:
            # symbol = f"\033[4m{symbol}!\033[0m"
            symbol_suffix += "!"
            if self._solver_terminated:
                symbol_suffix += "!"
        symbol += symbol_suffix
        lit_suffix = self._get_lit_suffix()
        symbol += lit_suffix

        def format_operand(op):
            if not isinstance(op, ParameterOperatable):
                return str(op)
            op_out = op.compact_repr(context, use_name=use_name)
            if isinstance(op, Expression) and len(op.operands) > 1:
                op_out = f"({op_out})"
            return op_out

        formatted_operands = [format_operand(op) for op in self.operands]
        out = ""
        if style.placement == Expression.ReprStyle.Placement.PREFIX:
            if len(formatted_operands) == 1:
                out = f"{symbol}{formatted_operands[0]}"
            else:
                out = f"{symbol}({', '.join(formatted_operands)})"
        elif style.placement == Expression.ReprStyle.Placement.EMBRACE:
            out = f"{symbol}{', '.join(formatted_operands)}{style.symbol}"
        elif len(formatted_operands) == 0:
            out = f"{type(self).__name__}{symbol_suffix}()"
        elif style.placement == Expression.ReprStyle.Placement.POSTFIX:
            if len(formatted_operands) == 1:
                out = f"{formatted_operands[0]}{symbol}"
            else:
                out = f"({', '.join(formatted_operands)}){symbol}"
        elif len(formatted_operands) == 1:
            out = f"{type(self).__name__}{symbol_suffix}({formatted_operands[0]})"
        elif lit_suffix and len(formatted_operands) > 2:
            out = (
                f"{type(self).__name__}{symbol_suffix}{lit_suffix}"
                f"({', '.join(formatted_operands)})"
            )
        elif style.placement == Expression.ReprStyle.Placement.INFIX:
            symbol = f" {symbol} "
            out = f"{symbol.join(formatted_operands)}"
        elif style.placement == Expression.ReprStyle.Placement.INFIX_FIRST:
            if len(formatted_operands) == 2:
                out = f"{formatted_operands[0]} {symbol} {formatted_operands[1]}"
            else:
                out = (
                    f"{formatted_operands[0]}{symbol}("
                    f"{', '.join(formatted_operands[1:])})"
                )
        else:
            assert False
        assert out

        # out += self._get_lit_suffix()

        return out

    def __rich_repr__(self):
        if not self._is_setup:
            yield f"{type(self)}(not init)"
        else:
            yield self.compact_repr()


@abstract
class ConstrainableExpression(Expression):
    def __init__(self, *operands: ParameterOperatable.All):
        super().__init__(Boolean(), *operands)
        self.constrained: bool = False

        # TODO this should be done in solver, not here
        self._solver_terminated: bool = False
        """
        Flag marking to the solver that this predicate has been deduced to True.
        Differs from alias in the sense that we can guarantee that the predicate is
        True, while alias only marks that the predicate shall be True.
        """

    def _constrain(self, constraint: "Predicate"):
        constraint.constrain()

    # shortcuts
    def constrain(self):
        self.constrained = True
        return self

    # should be eager, in the sense that, if the outcome is known, the callable is
    # called immediately, without storing an expression
    # we must force a value (at the end of solving at the least)
    def if_then_else(
        self,
        if_true: Callable[[], Any],
        if_false: Callable[[], Any],
        preference: bool | None = None,
    ):
        return IfThenElse(self, if_true, if_false, preference)

    @staticmethod
    def operation_switch_case_implications(
        cases: Iterable[tuple["ConstrainableExpression", "ConstrainableExpression"]],
    ) -> "ConstrainableExpression":
        return And(*(case.operation_implies(impl) for case, impl in cases))


@abstract
class Arithmetic(Expression):
    def __init__(self, *operands: ParameterOperatable.NumberLike):
        # HasUnit attr
        self.units: Unit = cast(Unit, None)  # checked in postinit

        super().__init__(Numbers(), *operands)
        types = (
            int,
            float,
            Quantity,
            Unit,
            Parameter,
            Arithmetic,
            Quantity_Interval,
            Quantity_Interval_Disjoint,
        )
        invalid_operands = [op for op in operands if not isinstance(op, types)]
        if invalid_operands:
            raise ValueError(
                "operands must be int, float, Quantity, Unit, Parameter, Arithmetic"
                ", Quantity_Interval, or Quantity_Interval_Disjoint"
                f", got {invalid_operands} with types"
                f" ([{', '.join(type(op).__name__ for op in invalid_operands)}])"
            )
        if any(
            not isinstance(param.domain, (Numbers, ESeries))
            for param in operands
            if isinstance(param, Parameter)
        ):
            raise ValueError("parameters must have domain Numbers or ESeries")

        # FIXME: convert to Quantity

        # TODO enforce
        self.operands = cast(tuple[ParameterOperatable.NumberLike, ...], operands)

    def __postinit__(self):
        assert self.units is not None


@abstract
class Additive(Arithmetic):
    def __init__(self, *operands):
        super().__init__(*operands)
        if not operands:
            self.units = dimensionless
        else:
            self.units = assert_compatible_units(operands)

    @staticmethod
    def sum(operands: Sequence[ParameterOperatable.NumberLike]) -> "Additive":
        # Else assert not correct
        if not len(operands):
            raise ValueError("at least one operand is required")
        return cast_assert(Additive, sum(operands))


class Add(Additive):
    REPR_STYLE = Additive.ReprStyle(
        symbol="+",
        placement=Additive.ReprStyle.Placement.INFIX,
    )

    def __init__(self, *operands):
        super().__init__(*operands)

    def has_implicit_constraint(self) -> bool:
        return False


class Subtract(Additive):
    REPR_STYLE = Additive.ReprStyle(
        symbol="-",
        placement=Additive.ReprStyle.Placement.INFIX,
    )

    def __init__(self, minuend, *subtrahend):
        super().__init__(minuend, *subtrahend)

    def has_implicit_constraint(self) -> bool:
        return False


class Multiply(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="*",
        placement=Arithmetic.ReprStyle.Placement.INFIX,
    )

    def __init__(self, *operands):
        super().__init__(*operands)
        units = [HasUnit.get_units_or_dimensionless(op) for op in operands]
        self.units = dimensionless
        for u in units:
            self.units = cast_assert(Unit, self.units * u)

    def has_implicit_constraint(self) -> bool:
        return False


class Divide(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="/",
        placement=Arithmetic.ReprStyle.Placement.INFIX,
    )

    def __init__(self, numerator, *denominator):
        super().__init__(numerator, *denominator)

        frac_unit = dimensionless
        for d in denominator:
            frac_unit = cast_assert(
                Unit, frac_unit * HasUnit.get_units_or_dimensionless(d)
            )

        self.units = cast_assert(
            Unit, HasUnit.get_units_or_dimensionless(numerator) / frac_unit
        )

    def has_implicit_constraint(self) -> bool:
        return True  # denominator not zero


class Sqrt(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="√",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, operand):
        super().__init__(operand)
        self.units = operand.units**0.5

    def has_implicit_constraint(self) -> bool:
        return True  # non-negative


class Power(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="^",
        placement=Arithmetic.ReprStyle.Placement.INFIX_FIRST,
    )

    def __init__(self, base, exponent):
        super().__init__(base, exponent)

        exp_unit = HasUnit.get_units_or_dimensionless(exponent)
        if not exp_unit.is_compatible_with(dimensionless):
            raise UnitCompatibilityError(
                "exponent must have dimensionless unit",
                incompatible_items=[exponent],
            )
        base_unit = HasUnit.get_units_or_dimensionless(base)
        units = dimensionless
        if not base_unit.is_compatible_with(dimensionless):
            if isinstance(exponent, ParameterOperatable):
                raise NotImplementedError(
                    "exponent must be a literal for base with unit"
                )
            exp_val = Quantity_Interval_Disjoint.from_value(exponent)
            if exp_val.min_elem != exp_val.max_elem:
                raise ValueError(
                    "exponent must be a single value for non-dimensionless base"
                )
            units = base_unit**exp_val.min_elem.magnitude
        assert isinstance(units, Unit)
        self.units = units


class Log(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="log",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, operand):
        super().__init__(operand)
        unit = HasUnit.get_units_or_dimensionless(operand)
        if not unit.is_compatible_with(dimensionless):
            raise ValueError("operand must have dimensionless unit")
        self.units = unit

        # TODO, be careful makes solver run forever like this
        # implicit constraint
        # if isinstance(operand, ParameterOperatable):
        #    operand.constrain_ge(lit(0))
        # elif Quantity_Interval_Disjoint.from_value(operand).min_elem < 0:
        #    raise ValueError("operand must be non-negative")

    def has_implicit_constraint(self) -> bool:
        return True  # non-negative


class Sin(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="sin",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, operand):
        super().__init__(operand)
        unit = HasUnit.get_units_or_dimensionless(operand)
        if not unit.is_compatible_with(dimensionless):
            raise ValueError("operand must have dimensionless unit")
        self.units = unit

    def has_implicit_constraint(self) -> bool:
        return False


class Cos(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="cos",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, operand):
        super().__init__(operand)
        unit = HasUnit.get_units_or_dimensionless(operand)
        if not unit.is_compatible_with(dimensionless):
            raise ValueError("operand must have dimensionless unit")
        self.units = unit

    def has_implicit_constraint(self) -> bool:
        return False


class Abs(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="|",
        placement=Arithmetic.ReprStyle.Placement.EMBRACE,
    )

    def __init__(self, operand):
        super().__init__(operand)
        self.units = HasUnit.get_units_or_dimensionless(operand)

    def has_implicit_constraint(self) -> bool:
        return False


class Round(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="round",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, operand):
        super().__init__(operand)
        self.units = HasUnit.get_units_or_dimensionless(operand)

    def has_implicit_constraint(self) -> bool:
        return False


class Floor(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="⌊",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, operand):
        super().__init__(operand)
        self.units = HasUnit.get_units_or_dimensionless(operand)

    def has_implicit_constraint(self) -> bool:
        return False


class Ceil(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="⌈",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, operand):
        super().__init__(operand)
        self.units = HasUnit.get_units_or_dimensionless(operand)

    def has_implicit_constraint(self) -> bool:
        return False


class Min(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="min",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, *operands):
        super().__init__(*operands)
        self.units = assert_compatible_units(operands)

    def has_implicit_constraint(self) -> bool:
        return False


class Max(Arithmetic):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="max",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, *operands):
        super().__init__(*operands)
        self.units = assert_compatible_units(operands)

    def has_implicit_constraint(self) -> bool:
        return False


@abstract
class Functional(Arithmetic):
    def __init__(self, function: "ParameterOperatable", variable: "Parameter"):
        super().__init__(function, variable)


class Integrate(Functional):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="∫",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, function: "ParameterOperatable", variable: "Parameter"):
        super().__init__(function, variable)
        # TODO units

    def has_implicit_constraint(self) -> bool:
        return False


class Differentiate(Functional):
    REPR_STYLE = Arithmetic.ReprStyle(
        symbol="d",
        placement=Arithmetic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, function: "ParameterOperatable", variable: "Parameter"):
        super().__init__(function, variable)
        # TODO units

    def has_implicit_constraint(self) -> bool:
        return False


class Logic(ConstrainableExpression):
    def __init__(self, *operands):
        super().__init__(*operands)
        types = bool, BoolSet, Parameter, Logic, Predicate
        if any(not isinstance(op, types) for op in operands):
            raise ValueError("operands must be bool, Parameter, Logic, or Predicate")
        if any(
            not isinstance(param.domain, Boolean)
            or not param.units.is_compatible_with(dimensionless)
            for param in operands
            if isinstance(param, Parameter)
        ):
            raise ValueError("parameters must have domain Boolean without a unit")


class And(Logic):
    REPR_STYLE = Logic.ReprStyle(
        symbol="∧",
        placement=Logic.ReprStyle.Placement.INFIX,
    )


class Or(Logic):
    REPR_STYLE = Logic.ReprStyle(
        symbol="∨",
        placement=Logic.ReprStyle.Placement.INFIX,
    )


class Not(Logic):
    REPR_STYLE = Logic.ReprStyle(
        symbol="¬",
        placement=Logic.ReprStyle.Placement.PREFIX,
    )

    def __init__(self, operand):
        super().__init__(operand)


class Xor(Logic):
    REPR_STYLE = Logic.ReprStyle(
        symbol="⊕",
        placement=Logic.ReprStyle.Placement.INFIX,
    )

    def __init__(self, *operands):
        super().__init__(*operands)


class Implies(Logic):
    REPR_STYLE = Logic.ReprStyle(
        symbol="⇒",
        placement=Logic.ReprStyle.Placement.INFIX_FIRST,
    )

    def __init__(self, condition, implication):
        super().__init__(condition, implication)


class IfThenElse(Expression):
    def __init__(
        self,
        condition: ConstrainableExpression,
        if_true: Callable[[], None] | None = None,
        if_false: Callable[[], None] | None = None,
        preference: bool | None = None,
    ):
        # FIXME domain
        super().__init__(None, condition)

        # TODO a bit hacky
        self.non_operands = [
            if_true or (lambda: None),
            if_false or (lambda: None),
            preference,
        ]

        # TODO actually implement this
        if preference is not None:
            if preference:
                condition.constrain()
                if if_true:
                    if_true()
            else:
                condition.operation_not().constrain()
                if if_false:
                    if_false()

    @property
    def if_true(self) -> Callable[[], None]:
        return self.non_operands[0]

    @property
    def if_false(self) -> Callable[[], None]:
        return self.non_operands[1]

    @property
    def preference(self) -> bool | None:
        return self.non_operands[2]


class Setic(Expression):
    def __init__(self, *operands):
        # FIXME domain
        super().__init__(None, *operands)

        if len(operands) == 0:
            raise ValueError("Setic must have at least one operand")
        # types = (Parameter, ParameterOperatable.Sets)
        # if any(not isinstance(op, types) for op in operands):
        #    raise ValueError("operands must be Parameter or Set")
        units = [HasUnit.get_units_or_dimensionless(op) for op in operands]
        self.units = units[0]
        for u in units[1:]:
            if not self.units.is_compatible_with(u):
                raise ValueError("all operands must have compatible units")
        # TODO domain?


class Union(Setic):
    REPR_STYLE = Setic.ReprStyle(
        symbol="∪",
        placement=Setic.ReprStyle.Placement.INFIX,
    )


class Intersection(Setic):
    REPR_STYLE = Setic.ReprStyle(
        symbol="∩",
        placement=Setic.ReprStyle.Placement.INFIX,
    )


class Difference(Setic):
    REPR_STYLE = Setic.ReprStyle(
        symbol="−",
        placement=Setic.ReprStyle.Placement.INFIX,
    )

    def __init__(self, minuend, *subtrahends):
        super().__init__(minuend, *subtrahends)


class SymmetricDifference(Setic):
    REPR_STYLE = Setic.ReprStyle(
        symbol="△",
        placement=Setic.ReprStyle.Placement.INFIX,
    )

    def __init__(self, left, right):
        super().__init__(left, right)


class Domain:
    @staticmethod
    def get_shared_domain(*domains: "Domain") -> "Domain":
        if len(domains) == 0:
            raise ValueError("No domains provided")
        if len(domains) == 1:
            return domains[0]
        one = domains[0]
        two = domains[1]
        match one:
            case Boolean():
                if not isinstance(two, Boolean):
                    raise ValueError(
                        "Boolean domain cannot be mixed with other domains"
                    )
                shared = Boolean()
            case EnumDomain():
                if not isinstance(two, EnumDomain):
                    raise ValueError("Enum domain cannot be mixed with other domains")
                if one.enum_t != two.enum_t:
                    raise ValueError("Enum domains must be of the same type")
                shared = EnumDomain(one.enum_t)
            case Numbers():
                if not isinstance(two, Numbers):
                    raise ValueError(
                        "Numbers domain cannot be mixed with other domains"
                    )
                if (
                    isinstance(one, ESeries)
                    and isinstance(two, ESeries)
                    and one.series == two.series
                ):
                    shared = ESeries(one.series)
                else:
                    shared = Numbers(
                        negative=one.negative and two.negative,
                        zero_allowed=one.zero_allowed and two.zero_allowed,
                        integer=one.integer or two.integer,
                    )
            case _:
                raise ValueError("Unsupported domain")

        if len(domains) == 2:
            return shared
        return Domain.get_shared_domain(shared, *domains[2:])

    def unbounded(self, param: "Parameter") -> P_Set: ...

    def __repr__(self):
        # TODO make more informative
        return f"{type(self).__name__}()"


class Numbers(Domain):
    def __init__(
        self, *, negative: bool = True, zero_allowed: bool = True, integer: bool = False
    ) -> None:
        super().__init__()
        self.negative = negative
        self.zero_allowed = zero_allowed
        self.integer = integer

    @override
    def unbounded(self, param: "Parameter") -> Quantity_Interval_Disjoint:
        if self.integer:
            # TODO
            _max = int(1e15)
            _min = -_max if self.negative else 0
            return Quantity_Interval_Disjoint.from_value(
                Quantity_Interval(
                    min=quantity(_min, param.units),
                    max=quantity(_max, param.units),
                    units=param.units,
                )
            )
        if not self.zero_allowed:
            raise NotImplementedError("Non-zero unbounded not implemented")
        if not self.negative:
            return Quantity_Interval_Disjoint.from_value(
                Quantity_Interval(
                    min=quantity(0, param.units), max=None, units=param.units
                )
            )
        return Quantity_Interval_Disjoint.unbounded(param.units)


class ESeries(Numbers):
    class SeriesType(Enum):
        E6 = auto()
        E12 = auto()
        E24 = auto()
        E48 = auto()
        E96 = auto()
        E192 = auto()

    def __init__(self, series: SeriesType):
        super().__init__(negative=False, zero_allowed=False, integer=False)
        self.series = series


class Boolean(Domain):
    @override
    def unbounded(self, param: "Parameter") -> BoolSet:
        return BoolSet.unbounded()


class EnumDomain(Domain):
    def __init__(self, enum_t: type[Enum]):
        super().__init__()
        self.enum_t = enum_t

    @override
    def unbounded(self, param: "Parameter") -> EnumSet:
        return EnumSet.unbounded(self.enum_t)


class Predicate(ConstrainableExpression):
    def __init__(self, left, right):
        super().__init__(left, right)
        assert_compatible_units(self.operands)

    def __bool__(self):
        raise ValueError("Predicate cannot be converted to bool")


class NumericPredicate(Predicate):
    def __init__(self, left, right):
        super().__init__(left, right)

        for op in self.operands:
            if isinstance(op, Parameter) and not isinstance(
                op.domain, (Numbers, ESeries)
            ):
                raise ValueError(
                    "operand must have domain Numbers or ESeries,"
                    f" not {type(op.domain)}"
                )


class LessThan(NumericPredicate):
    REPR_STYLE = NumericPredicate.ReprStyle(
        symbol="<",
        placement=NumericPredicate.ReprStyle.Placement.INFIX_FIRST,
    )

    def __init__(self, left, right):
        # TODO we might allow it for integer domains at some point
        raise NotImplementedError(
            "'<' not supported, you very likely want to use '<=' instead"
        )
        super().__init__(left, right)


class GreaterThan(NumericPredicate):
    REPR_STYLE = NumericPredicate.ReprStyle(
        symbol=">",
        placement=NumericPredicate.ReprStyle.Placement.INFIX_FIRST,
    )


class LessOrEqual(NumericPredicate):
    REPR_STYLE = NumericPredicate.ReprStyle(
        symbol="≤",
        placement=NumericPredicate.ReprStyle.Placement.INFIX_FIRST,
    )


class GreaterOrEqual(NumericPredicate):
    REPR_STYLE = NumericPredicate.ReprStyle(
        symbol="≥",
        placement=NumericPredicate.ReprStyle.Placement.INFIX_FIRST,
    )


class NotEqual(NumericPredicate):
    REPR_STYLE = NumericPredicate.ReprStyle(
        symbol="≠",
        placement=NumericPredicate.ReprStyle.Placement.INFIX_FIRST,
    )


class SeticPredicate(Predicate):
    def __init__(self, left, right):
        super().__init__(left, right)
        # types = ParameterOperatable, P_Set
        # TODO
        # if any(not isinstance(op, types) for op in self.operands):
        #    raise ValueError("operands must be Parameter or Set")
        units = [HasUnit.get_units_or_dimensionless(op) for op in self.operands]
        for u in units[1:]:
            if not units[0].is_compatible_with(u):
                raise ValueError("all operands must have compatible units")
        # TODO domain?


class IsSubset(SeticPredicate):
    REPR_STYLE = SeticPredicate.ReprStyle(
        symbol="⊆",
        placement=SeticPredicate.ReprStyle.Placement.INFIX_FIRST,
    )


class IsSuperset(SeticPredicate):
    REPR_STYLE = SeticPredicate.ReprStyle(
        symbol="⊇",
        placement=SeticPredicate.ReprStyle.Placement.INFIX_FIRST,
    )


class Cardinality(SeticPredicate):
    REPR_STYLE = SeticPredicate.ReprStyle(
        # TODO
        symbol="||",
        placement=SeticPredicate.ReprStyle.Placement.PREFIX,
    )

    def __init__(
        self, set: ParameterOperatable.Sets, cardinality: ParameterOperatable.NumberLike
    ):
        super().__init__(set, cardinality)


class Is(Predicate):
    REPR_STYLE = Predicate.ReprStyle(
        symbol="is",
        placement=Predicate.ReprStyle.Placement.INFIX_FIRST,
    )

    def __init__(self, left, right):
        super().__init__(left, right)


# TODO rename?
class R(Namespace):
    """
    Namespace holding Expressions, Domains and Predicates for Parameters.
    R = paRameters
    """

    class Predicates(Namespace):
        class Element(Namespace):
            LT = LessThan
            GT = GreaterThan
            LE = LessOrEqual
            GE = GreaterOrEqual
            NE = NotEqual

        class Set(Namespace):
            IS_SUBSET = IsSubset
            IS_SUPERSET = IsSuperset

    class Domains(Namespace):
        class ESeries(Namespace):
            E6 = lambda: ESeries(ESeries.SeriesType.E6)  # noqa: E731
            E12 = lambda: ESeries(ESeries.SeriesType.E12)  # noqa: E731
            E24 = lambda: ESeries(ESeries.SeriesType.E24)  # noqa: E731
            E48 = lambda: ESeries(ESeries.SeriesType.E48)  # noqa: E731
            E96 = lambda: ESeries(ESeries.SeriesType.E96)  # noqa: E731
            E192 = lambda: ESeries(ESeries.SeriesType.E192)  # noqa: E731

        class Numbers(Namespace):
            REAL = Numbers
            NATURAL = lambda: Numbers(integer=True, negative=False)  # noqa: E731

        BOOL = Boolean
        ENUM = EnumDomain

    class Expressions(Namespace):
        class Arithmetic(Namespace):
            ADD = Add
            SUBTRACT = Subtract
            MULTIPLY = Multiply
            DIVIDE = Divide
            POWER = Power
            LOG = Log
            SQRT = Sqrt
            LOG = Log
            ABS = Abs
            FLOOR = Floor
            CEIL = Ceil
            ROUND = Round
            SIN = Sin
            COS = Cos

        class Logic(Namespace):
            AND = And
            OR = Or
            NOT = Not
            XOR = Xor
            IMPLIES = Implies

        class Set(Namespace):
            UNION = Union
            INTERSECTION = Intersection
            DIFFERENCE = Difference
            SYMMETRIC_DIFFERENCE = SymmetricDifference


class Parameter(ParameterOperatable):
    class TraitT(Trait): ...

    def __init__(
        self,
        *,
        units: Unit | Quantity = dimensionless,
        # hard constraints
        within: Quantity_Interval_Disjoint | Quantity_Interval | None = None,
        domain: Domain = Numbers(negative=False),
        # soft constraints
        soft_set: Quantity_Interval_Disjoint | Quantity_Interval | None = None,
        guess: Quantity
        | int
        | float
        | None = None,  # TODO actually allowed to be anything from domain
        tolerance_guess: float | None = None,
        # hints
        likely_constrained: bool = False,  # TODO rename expect_constraits or similiar
    ):
        super().__init__()
        if within is not None and not within.units.is_compatible_with(units):
            raise ValueError("incompatible units")

        if isinstance(within, Quantity_Interval):
            within = Quantity_Interval_Disjoint(within)

        if isinstance(soft_set, Quantity_Interval):
            soft_set = Quantity_Interval_Disjoint(soft_set)

        if not isinstance(units, Unit):
            raise TypeError("units must be a Unit")
        self.units = units
        self.within = within
        self._domain = domain
        self.soft_set = soft_set
        self.guess = guess
        self.tolerance_guess = tolerance_guess
        self.likely_constrained = likely_constrained

    # Type forwards
    type All = ParameterOperatable.All
    type NumberLike = ParameterOperatable.NumberLike
    type Sets = ParameterOperatable.Sets
    type BooleanLike = ParameterOperatable.BooleanLike
    type Number = ParameterOperatable.Number

    @property
    def domain(self) -> Domain:
        return self._domain

    def has_implicit_constraint(self) -> bool:
        return False

    def has_implicit_constraints_recursive(self) -> bool:
        return False

    @once
    def compact_repr(
        self,
        context: ParameterOperatable.ReprContext | None = None,
        use_name: bool = False,
    ) -> str:
        """
        Unit only printed if not dimensionless.

        Letters:
        ```
        A-Z, a-z, α-ω
        A₁-Z₁, a₁-z₁, α₁-ω₁
        A₂-Z₂, a₂-z₂, α₂-ω₂
        ...
        ```
        """

        def param_id_to_human_str(param_id: int) -> str:
            assert isinstance(param_id, int)
            alphabets = [("A", 26), ("a", 26), ("α", 25)]
            alphabet = [
                chr(ord(start_char) + i)
                for start_char, char_cnt in alphabets
                for i in range(char_cnt)
            ]

            def int_to_subscript(i: int) -> str:
                if i == 0:
                    return ""
                _str = str(i)
                return "".join(chr(ord("₀") + ord(c) - ord("0")) for c in _str)

            return alphabet[param_id % len(alphabet)] + int_to_subscript(
                param_id // len(alphabet)
            )

        if use_name and self.get_parent() is not None:
            letter = self.get_full_name()
        else:
            if context is None:
                context = ParameterOperatable.ReprContext()
            if self not in context.variable_mapping.mapping:
                next_id = context.variable_mapping.next_id
                context.variable_mapping.mapping[self] = next_id
                context.variable_mapping.next_id += 1
            letter = param_id_to_human_str(context.variable_mapping.mapping[self])

        unitstr = f" {self.units}" if self.units != dimensionless else ""

        out = f"{letter}{unitstr}"
        out += self._get_lit_suffix()

        return out

    def domain_set(self) -> P_Set:
        return self.domain.unbounded(self)

    def get_last_known_deduced_superset(self, solver: "Solver") -> P_Set | None:
        as_literal = solver.inspect_get_known_supersets(self)
        return None if as_literal == self.domain_set() else as_literal


p_field = f_field(Parameter)

# Canonical ----------------------------------------------------------------------------
CanonicalNumericExpression = Add | Multiply | Power | Round | Abs | Sin | Log
CanonicalLogicExpression = Or | Not
CanonicalSeticExpression = Intersection | Union | SymmetricDifference
CanonicalPredicate = GreaterOrEqual | IsSubset | Is | GreaterThan

CanonicalConstrainableExpression = CanonicalLogicExpression | CanonicalPredicate

CanonicalExpression = (
    CanonicalNumericExpression
    | CanonicalLogicExpression
    | CanonicalSeticExpression
    | CanonicalPredicate
)
CanonicalNumber = Quantity_Interval_Disjoint | Quantity_Set_Discrete
CanonicalBoolean = BoolSet
CanonicalEnum = P_Set[Enum]
# TODO Canonical set?

CanonicalLiteral = CanonicalNumber | CanonicalBoolean | CanonicalEnum
CanonicalOperable = CanonicalExpression | Parameter
CanonicalAll = CanonicalOperable | CanonicalLiteral

# --------------------------------------------------------------------------------------

Reflexive = Is | IsSubset | GreaterOrEqual
IdempotentExpression = Abs | Round
IdempotentOperands = Or | Union | Intersection
Commutative = Add | Multiply | Or | Union | Intersection | SymmetricDifference | Is
UnaryIdentity = Add | Multiply | Or | Union | Intersection
FullyAssociative = Add | Multiply | Or | Union | Intersection
Associative = FullyAssociative
Involutory = Not


# python help --------------------------------------------------------------------------
CanonicalExpressionR = (
    Add,
    Multiply,
    Power,
    Round,
    Abs,
    Sin,
    Log,
    Or,
    Not,
    Intersection,
    Union,
    SymmetricDifference,
    Is,
    GreaterOrEqual,
    GreaterThan,
    IsSubset,
)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class ResistorVoltageDivider(Module):
    """
    A voltage divider using two resistors.
    node[0] ~ resistor[1] ~ node[1] ~ resistor[2] ~ node[2]
    power.hv ~ node[0]
    power.lv ~ node[2]
    output.line ~ node[1]
    output.reference.lv ~ node[2]
    """

    # External interfaces
    power: F.ElectricPower
    output: F.ElectricSignal

    # Components
    r_bottom: F.Resistor
    r_top: F.Resistor

    # Variables
    v_in = L.p_field(units=P.V)
    v_out = L.p_field(units=P.V)
    max_current = L.p_field(units=P.A)
    total_resistance = L.p_field(units=P.Ω)
    ratio = L.p_field(units=P.dimensionless)

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.power.hv, self.output.line)

    def __preinit__(self):
        # Connections
        self.power.hv.connect_via(
            [self.r_top, self.output.line, self.r_bottom], self.power.lv
        )

        # Short variables
        ratio = self.ratio
        r_top = self.r_top.resistance
        r_bottom = self.r_bottom.resistance
        r_total = self.total_resistance
        v_out = self.v_out
        v_in = self.v_in
        max_current = self.max_current

        # Link interface voltages
        v_out.alias_is(self.output.reference.voltage)
        v_in.alias_is(self.power.voltage)

        # Equations
        r_top.alias_is((v_in / max_current) - r_bottom)
        r_bottom.alias_is((v_in / max_current) - r_top)
        r_top.alias_is((v_in - v_out) / max_current)
        r_bottom.alias_is(v_out / max_current)
        r_bottom.alias_is(r_total * ratio)
        r_top.alias_is(r_total * (1 - ratio))
        r_bottom.alias_is(v_out * r_top / (v_in - v_out))
        r_top.alias_is(r_bottom * (v_in / v_out - 1))

        # Calculate outputs
        r_total.alias_is(r_top + r_bottom)
        v_out.alias_is(v_in * r_bottom / r_total)
        v_out.alias_is(v_in * ratio)
        max_current.alias_is(v_in / r_total)
        ratio.alias_is(r_bottom / r_total)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import collections.abc
import hashlib
import importlib.util
import inspect
import itertools
import json
import logging
import os
import select
import shutil
import stat
import subprocess
import sys
import time
import uuid
from abc import abstractmethod
from collections import defaultdict, deque
from contextlib import contextmanager
from dataclasses import dataclass, fields
from enum import Enum, StrEnum
from functools import wraps
from genericpath import commonprefix
from importlib.metadata import Distribution
from itertools import chain, pairwise
from json import JSONEncoder
from pathlib import Path
from textwrap import indent
from types import ModuleType
from typing import (
    Any,
    Callable,
    Concatenate,
    Container,
    Generator,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Protocol,
    Self,
    Sequence,
    SupportsFloat,
    SupportsInt,
    Type,
    TypeGuard,
    cast,
    get_origin,
    get_type_hints,
    overload,
    override,
)

import psutil

logger = logging.getLogger(__name__)


class Serializable(Protocol):
    def serialize(self) -> dict: ...

    @classmethod
    def deserialize(cls, data: dict) -> Self: ...


class SerializableJSONEncoder(JSONEncoder):
    def default(self, o: Serializable | None):
        return None if o is None else o.serialize()


class lazy:
    def __init__(self, expr):
        self.expr = expr

    def __str__(self):
        return str(self.expr())

    def __repr__(self):
        return repr(self.expr())


def kw2dict(**kw):
    return dict(kw)


class hashable_dict:
    def __init__(self, obj: dict):
        self.obj = obj

    def __hash__(self):
        return hash(sum(map(hash, self.obj.items())))

    def __repr__(self):
        return "{}({})".format(type(self), repr(self.obj))

    def __eq__(self, other):
        return hash(self) == hash(other)


def unique[T, U](it: Iterable[T], key: Callable[[T], U]) -> list[T]:
    seen: list[U] = []
    out: list[T] = []
    for i in it:
        v = key(i)
        if v in seen:
            continue
        seen.append(v)
        out.append(i)
    return out


def unique_ref[T](it: Iterable[T]) -> list[T]:
    return unique(it, id)


def duplicates[T, U](
    it: Iterable[T], key: Callable[[T], U], by_eq: bool = False
) -> dict[U, list[T]]:
    if by_eq:
        return {
            k: uv
            for k, v in groupby(it, key).items()
            if len(uv := unique(v, key=lambda x: x)) > 1
        }
    else:
        return {k: v for k, v in groupby(it, key).items() if len(v) > 1}


def get_dict(obj, key, default):
    if key not in obj:
        obj[key] = default()

    return obj[key]


def flatten(obj: Iterable, depth=1) -> List:
    if depth == 0:
        return list(obj)
    if not isinstance(obj, Iterable):
        return [obj]
    return [nested for top in obj for nested in flatten(top, depth=depth - 1)]


def get_key[T, U](haystack: dict[T, U], needle: U) -> T:
    return find(haystack.items(), lambda x: x[1] == needle)[0]


class KeyErrorNotFound(KeyError): ...


class KeyErrorAmbiguous[T](KeyError):
    def __init__(self, duplicates: list[T], *args: object) -> None:
        super().__init__(*args)
        self.duplicates = duplicates

    def __str__(self):
        return f"KeyErrorAmbiguous: {self.duplicates}"


def find[T](haystack: Iterable[T], needle: Callable[[T], Any] | None = None) -> T:
    if needle is None:
        needle = lambda x: x is not None  # noqa: E731
    results = [x for x in haystack if needle(x)]
    if not results:
        raise KeyErrorNotFound()
    if len(results) != 1:
        raise KeyErrorAmbiguous(results)
    return results[0]


def find_or[T](
    haystack: Iterable[T],
    needle: Callable[[T], bool],
    default: T,
    default_multi: Callable[[list[T]], T] | None = None,
) -> T:
    try:
        return find(haystack, needle)
    except KeyErrorNotFound:
        return default
    except KeyErrorAmbiguous as e:
        if default_multi is not None:
            return default_multi(e.duplicates)
        raise


def groupby[T, U](
    it: Iterable[T], key: Callable[[T], U], only_multi: bool = False
) -> dict[U, list[T]]:
    out = defaultdict(list)
    for i in it:
        out[key(i)].append(i)
    if only_multi:
        return {k: v for k, v in out.items() if len(v) > 1}
    return out


def nested_enumerate(it: Iterable) -> list[tuple[list[int], Any]]:
    out: list[tuple[list[int], Any]] = []
    for i, obj in enumerate(it):
        if not isinstance(obj, Iterable):
            out.append(([i], obj))
            continue
        for j, _obj in nested_enumerate(obj):
            out.append(([i] + j, _obj))

    return out


class NotifiesOnPropertyChange(object):
    def __init__(self, callback) -> None:
        self._callback = callback

        # TODO dir -> vars?
        for name in dir(self):
            self._callback(name, getattr(self, name))

    def __setattr__(self, __name, __value) -> None:
        super().__setattr__(__name, __value)

        # before init
        if hasattr(self, "_callback"):
            self._callback(__name, __value)


class _wrapper[T, P](NotifiesOnPropertyChange):
    @abstractmethod
    def __init__(self, parent: P) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_all(self) -> list[T]:
        raise NotImplementedError

    @abstractmethod
    def handle_add(self, name: str, obj: T):
        raise NotImplementedError

    @abstractmethod
    def get_parent(self) -> P:
        raise NotImplementedError

    @abstractmethod
    def extend_list(self, list_name: str, *objs: T) -> None:
        raise NotImplementedError


def Holder[T, P](_type: Type[T], _ptype: Type[P]) -> Type[_wrapper[T, P]]:
    class __wrapper[_T, _P](_wrapper[_T, _P]):
        def __init__(self, parent: P) -> None:
            self._list: list[T] = []
            self._type = _type
            self._parent: P = parent

            NotifiesOnPropertyChange.__init__(self, self._callback)

        def _callback(self, name: str, value: Any):
            if name.startswith("_"):
                return

            if callable(value):
                return

            if isinstance(value, self._type):
                self._list.append(value)
                self.handle_add(name, value)
                return

            if isinstance(value, dict):
                value = value.values()

            if isinstance(value, Iterable):
                e_objs = nested_enumerate(value)
                objs = [x[1] for x in e_objs]
                assert all(map(lambda x: isinstance(x, self._type), objs))

                self._list += objs
                for i_list, instance in e_objs:
                    i_acc = "".join(f"[{i}]" for i in i_list)
                    self.handle_add(f"{name}{i_acc}", instance)
                return

            raise Exception(
                f"Invalid property added for {name=} {value=} of type {type(value)},"
                + f"expected {_type} or iterable thereof"
            )

        def extend_list(self, list_name: str, *objs: T) -> None:
            if not hasattr(self, list_name):
                setattr(self, list_name, [])
            for obj in objs:
                # emulate property setter
                list_obj = getattr(self, list_name)
                idx = len(list_obj)
                list_obj.append(obj)
                self._list.append(obj)
                self.handle_add(f"{list_name}[{idx}]", obj)

        def get_all(self) -> list[T]:
            # check for illegal list modifications
            for name in sorted(dir(self)):
                value = getattr(self, name)
                if name.startswith("_"):
                    continue
                if callable(value):
                    continue
                if isinstance(value, self._type):
                    continue
                if isinstance(value, dict):
                    value = value.values()
                if isinstance(value, Iterable):
                    assert set(flatten(value, -1)).issubset(set(self._list))
                    continue

            return self._list

        def handle_add(self, name: str, obj: T) -> None: ...

        def get_parent(self) -> P:
            return self._parent

        def repr(self):
            return f"{type(self).__name__}({self._list})"

    return __wrapper[T, P]


def not_none(x):
    assert x is not None
    return x


@overload
def cast_assert[T](t: type[T], obj) -> T: ...


@overload
def cast_assert[T1, T2](t: tuple[type[T1], type[T2]], obj) -> T1 | T2: ...


@overload
def cast_assert[T1, T2, T3](
    t: tuple[type[T1], type[T2], type[T3]], obj
) -> T1 | T2 | T3: ...


@overload
def cast_assert[T1, T2, T3, T4](
    t: tuple[type[T1], type[T2], type[T3], type[T4]], obj
) -> T1 | T2 | T3 | T4: ...


def cast_assert(t, obj):
    """
    Assert that obj is an instance of type t and return it with proper type hints.
    t can be either a single type or a tuple of types.
    """
    assert isinstance(obj, t)
    return obj


def times[T](cnt: SupportsInt, lamb: Callable[[], T]) -> list[T]:
    return [lamb() for _ in range(int(cnt))]


def is_type_pair[T, U](
    param1: Any, param2: Any, type1: type[T], type2: type[U]
) -> Optional[tuple[T, U]]:
    o1 = get_origin(type1) or type1
    o2 = get_origin(type2) or type2
    if isinstance(param1, o1) and isinstance(param2, o2):
        return param1, param2
    if isinstance(param2, o1) and isinstance(param1, o2):
        return param2, param1
    return None


def is_type_set_subclasses(type_subclasses: set[type], types: set[type]) -> bool:
    hits = {t: any(issubclass(s, t) for s in type_subclasses) for t in types}
    return all(hits.values()) and all(
        any(issubclass(s, t) for t in types) for s in type_subclasses
    )


def round_str(value: SupportsFloat, n=8):
    """
    Round a float to n decimals and strip trailing zeros.
    """
    f = round(float(value), n)
    return str(f).rstrip("0").rstrip(".")


def _print_stack(stack) -> Iterable[str]:
    from rich.text import Text

    for frame_info in stack:
        frame = frame_info[0]
        if "venv" in frame_info.filename:
            continue
        if "faebryk" not in frame_info.filename:
            continue
        # if frame_info.function not in ["_connect_across_hierarchies"]:
        #    continue
        yield str(
            Text.assemble(
                (
                    f" Frame in {frame_info.filename} at line {frame_info.lineno}:",
                    "red",
                ),
                (f" {frame_info.function} ", "blue"),
            )
        )

        def pretty_val(value):
            if isinstance(value, dict):
                import pprint

                formatted = pprint.pformat(
                    {pretty_val(k): pretty_val(v) for k, v in value.items()},
                    indent=2,
                    width=120,
                )
                return ("\n" if len(value) > 1 else "") + indent(
                    str(Text(formatted)), " " * 4
                )
            elif isinstance(value, type):
                return f"<class {value.__name__}>"
            return str(value)

        for name, value in frame.f_locals.items():
            yield str(
                Text.assemble(
                    ("  ", ""),
                    (f"{name}", "green"),
                    (" = ", ""),
                    (pretty_val(value), ""),
                )
            )


def print_stack(stack):
    return "\n".join(_print_stack(stack))


# Get deepest values in nested dict:
def flatten_dict(d: dict):
    for k, v in d.items():
        if isinstance(v, dict):
            yield from flatten_dict(v)
        else:
            yield (k, v)


def split_recursive_stack(
    stack: Iterable[inspect.FrameInfo],
) -> tuple[list[inspect.FrameInfo], int, list[inspect.FrameInfo]]:
    """
    Handles RecursionError by splitting the stack into three parts:
    - recursion: the repeating part of the stack indicating the recursion.
    - stack_towards_recursion: the part of the stack after the recursion
        has been detected.

    :param stack: The stack obtained from inspect.stack()
    :return: tuple (recursion, recursion_depth, stack_towards_recursion)
    """

    def find_loop_len(sequence):
        for loop_len in range(1, len(sequence) // 2 + 1):
            if len(sequence) % loop_len:
                continue
            is_loop = True
            for i in range(0, len(sequence), loop_len):
                if sequence[i : i + loop_len] != sequence[:loop_len]:
                    is_loop = False
                    break
            if is_loop:
                return loop_len

        return 0

    def find_last_longest_most_frequent_looping_sequence_in_beginning(stack):
        stack = list(stack)

        loops = []

        # iterate over all possible beginnings
        for i in range(len(stack)):
            # iterate over all possible endings
            # try to maximize length of looping sequence
            for j in reversed(range(i + 1, len(stack) + 1)):
                # determine length of loop within this range
                loop_len = find_loop_len(stack[i:j])
                if loop_len:
                    # check if skipped beginning is partial loop
                    if stack[:i] != stack[j - i : j]:
                        continue
                    loops.append((i, j, loop_len))
                    continue

        # print(loops)
        max_loop = max(loops, key=lambda x: (x[1] - x[0], x[1]), default=None)
        return max_loop

    stack = list(stack)

    # Get the full stack representation as a list of strings
    full_stack = [f"{frame.filename}:{frame.positions}" for frame in stack]

    max_loop = find_last_longest_most_frequent_looping_sequence_in_beginning(full_stack)
    assert max_loop
    i, j, depth = max_loop

    return stack[i : i + depth], depth, stack[j:]


CACHED_RECUSION_ERRORS = set()


# TODO: now unused
# Consider splitting into three functions
# - A "lower_recursion_limit" contextmanager, should increase limit from current usage
# - A "better_recursion_error" function, should improve recursion error messaging
# - A separate composable "except" decorator which can be used to more generically
#   return something in case of a function raising an error
def try_avoid_endless_recursion(f: Callable[..., str]):
    import sys

    def _f_no_rec(*args, **kwargs):
        limit = sys.getrecursionlimit()
        target = 100
        sys.setrecursionlimit(target)
        try:
            return f(*args, **kwargs)
        except RecursionError:
            sys.setrecursionlimit(target + 1000)

            rec, depth, non_rec = split_recursive_stack(inspect.stack()[1:])
            recursion_error_str = indent(
                "\n".join(
                    [
                        f"{frame.filename}:{frame.lineno} {frame.code_context}"
                        for frame in rec
                    ]
                    + [f"... repeats {depth} times ..."]
                    + [
                        f"{frame.filename}:{frame.lineno} {frame.code_context}"
                        for frame in non_rec
                    ]
                ),
                "   ",
            )

            if recursion_error_str in CACHED_RECUSION_ERRORS:
                logger.error(
                    f"Recursion error: {f.__name__} {f.__code__.co_filename}:"
                    + f"{f.__code__.co_firstlineno}: DUPLICATE"
                )
            else:
                CACHED_RECUSION_ERRORS.add(recursion_error_str)
                logger.error(
                    f"Recursion error: {f.__name__} {f.__code__.co_filename}:"
                    + f"{f.__code__.co_firstlineno}"
                )
                logger.error(recursion_error_str)

            return "<RECURSION ERROR WHILE CONVERTING TO STR>"
        finally:
            sys.setrecursionlimit(limit)

    return _f_no_rec


def zip_non_locked[T, U](left: Iterable[T], right: Iterable[U]):
    # Theoretically supports any amount of iters,
    #  but for type hinting limit to two for now

    class _Iter[TS, US](Iterator[tuple[TS, US]]):
        class _NONDEFAULT: ...

        def __init__(self, args: list[Iterable]):
            self.iters = [iter(arg) for arg in args]
            self.stopped = False
            self.stepped = False
            self.values = [None for _ in self.iters]

        def next(self, i: int, default: Any = _NONDEFAULT):
            try:
                self.advance(i)
                return self.values[i]
            except StopIteration as e:
                self.stopped = True
                if default is not self._NONDEFAULT:
                    return default
                raise e

        def advance(self, i: int):
            self.values[i] = next(self.iters[i])
            self.stepped = True

        def advance_all(self):
            self.stepped = True
            try:
                self.values = [next(iter) for iter in self.iters]
            except StopIteration:
                self.stopped = True

        def __next__(self):
            if not self.stepped:
                self.advance_all()
            if self.stopped:
                raise StopIteration()
            self.stepped = False

            return tuple(self.values)

    return _Iter[T, U]([left, right])


def try_or[T](
    func: Callable[..., T],
    default: T | None = None,
    default_f: Callable[[Exception], T] | None = None,
    catch: type[Exception] | tuple[type[Exception], ...] = Exception,
) -> T:
    try:
        return func()
    except catch as e:
        if default_f is not None:
            default = default_f(e)
        return default


class SharedReference[T]:
    @dataclass
    class Resolution[U, S]:
        representative: S
        object: U
        old: U

    def __init__(self, object: T):
        self.object: T = object
        self.links: set[Self] = set([self])

    def link(self, other: Self):
        assert type(self) is type(other), f"{type(self)=} {type(other)=}"
        if self == other:
            return

        lhs, rhs = self, other
        old = rhs.object

        r_links = rhs.links
        for rhs_ in r_links:
            rhs_.object = lhs.object
            rhs_.links = lhs.links

        lhs.links.update(r_links)

        return self.Resolution(lhs, lhs.object, old)

    def set(self, obj: T):
        self.object = obj
        for link in self.links:
            link.object = obj

    def __call__(self) -> T:
        return self.object

    def __eq__(self, other: "SharedReference[T]"):
        return self.object is other.object and self.links is other.links

    def __hash__(self) -> int:
        return hash(id(self))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.object})"


def bfs_visit[T](
    neighbours: Callable[[list[T]], list[T]], roots: Iterable[T]
) -> set[T]:
    """
    Generic BFS (not depending on Graph)
    Returns all visited nodes.
    """
    open_path_queue: list[list[T]] = [[root] for root in roots]
    visited: set[T] = set(roots)

    while open_path_queue:
        open_path = open_path_queue.pop(0)

        for neighbour in neighbours(open_path):
            if neighbour not in visited:
                new_path = open_path + [neighbour]
                visited.add(neighbour)
                open_path_queue.append(new_path)

    return visited


class TwistArgs:
    def __init__(self, op: Callable) -> None:
        self.op = op

    def __call__(self, *args, **kwargs):
        return self.op(*reversed(args), **kwargs)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.op})"


class CallOnce[F: Callable]:
    def __init__(self, f: F) -> None:
        self.f = f
        self.called = False

    # TODO types
    def __call__(self, *args, **kwargs) -> Any:
        if self.called:
            return
        self.called = True
        return self.f(*args, **kwargs)


def at_exit(func: Callable, on_exception: bool = True):
    import atexit
    import sys

    f = CallOnce(func)

    atexit.register(f)
    hook = sys.excepthook
    if on_exception:
        sys.excepthook = lambda *args: (f(), hook(*args))

    # get main thread
    import threading

    mainthread = threading.main_thread()

    def wait_main():
        mainthread.join()
        f()

    threading.Thread(target=wait_main).start()

    return f


def lazy_construct(cls):
    """
    Careful: break deepcopy
    """
    old_init = cls.__init__

    def new_init(self, *args, **kwargs):
        self._init = False
        self._old_init = lambda: old_init(self, *args, **kwargs)

    def __getattr__(self, name: str, /):
        assert "_init" in self.__dict__
        if self._init:
            raise AttributeError(name)
        self._old_init()
        self._init = True
        return self.__getattribute__(name)

    cls.__init__ = new_init
    cls.__getattr__ = __getattr__
    return cls


# TODO figure out nicer way (with metaclass or decorator)
class LazyMixin:
    @property
    def is_init(self):
        return self.__dict__.get("_init", False)

    def force_init(self):
        if self.is_init:
            return
        self._old_init()
        self._init = True


class Lazy(LazyMixin):
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        lazy_construct(cls)


def once[T, **P](f: Callable[P, T]) -> Callable[P, T]:
    # TODO add flag for this optimization
    # might not be desirable if different instances with same hash
    # return same values here
    # check if f is a method with only self
    params = inspect.signature(f).parameters
    # optimization: if takes self, cache in instance (saves hash of instance)
    if "self" in params:
        name = f.__name__
        attr_name = f"_{name}_once"
        param_list = list(params)

        # optimization: if takes only self, no need for dict
        if len(param_list) == 1:

            def wrapper_single(self) -> Any:
                if not hasattr(self, attr_name):
                    setattr(self, attr_name, f(self))
                return getattr(self, attr_name)

            return wrapper_single

        # optimization: if takes self + args, use self as cache
        def wrapper_self(*args: P.args, **kwargs: P.kwargs) -> Any:
            self = args[0]
            lookup = (args[1:], tuple(kwargs.items()))
            if not hasattr(self, attr_name):
                setattr(self, attr_name, {})

            cache = getattr(self, attr_name)
            if lookup in cache:
                return cache[lookup]

            result = f(*args, **kwargs)
            cache[lookup] = result
            return result

        return wrapper_self

    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
        lookup = (args, tuple(kwargs.items()))
        if lookup in wrapper.cache:
            return wrapper.cache[lookup]

        result = f(*args, **kwargs)
        wrapper.cache[lookup] = result
        return result

    wrapper.cache = {}
    wrapper._is_once_wrapper = True
    return wrapper


def assert_once[T, O, **P](
    f: Callable[Concatenate[O, P], T],
) -> Callable[Concatenate[O, P], T]:
    def wrapper(obj: O, *args: P.args, **kwargs: P.kwargs) -> T:
        if not hasattr(obj, "_assert_once_called"):
            setattr(obj, "_assert_once_called", set())

        wrapper_set = getattr(obj, "_assert_once_called")

        if wrapper not in wrapper_set:
            wrapper_set.add(wrapper)
            return f(obj, *args, **kwargs)
        else:
            raise AssertionError(f"{f.__name__} called on {obj} more than once")

    return wrapper


def assert_once_global[T, **P](f: Callable[P, T]) -> Callable[P, T]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        if not wrapper.called:
            wrapper.called = True
            return f(*args, **kwargs)
        else:
            raise AssertionError("Function called more than once")

    wrapper.called = False
    return wrapper


class _ConfigFlagBase[T]:
    def __init__(self, name: str, default: T, descr: str = ""):
        self._name = name
        self.default = default
        self.descr = descr
        self._type: type[T] = type(default)
        self.value = self._get()
        self._has_been_read = False

    @property
    def name(self) -> str:
        return f"FBRK_{self._name}"

    def set(self, value: T, force: bool = False):
        if self._has_been_read and value != self.value and not force:
            raise ValueError(
                f"Can't write flag {self.name}"
                ", has already been read with different value"
            )
        self.value = value

    @property
    def raw_value(self) -> str | None:
        return os.getenv(self.name, None)

    def get(self) -> T:
        self._has_been_read = True
        return self.value

    def _get(self) -> T:
        raw_val = self.raw_value

        if raw_val is None:
            res = self.default
        else:
            res = self._convert(raw_val)

        if res != self.default:
            logger.warning(f"Config flag |{self.name}={res}|")

        return res

    def __hash__(self) -> int:
        return hash(self.name)

    @abstractmethod
    def _convert(self, raw_val: str) -> T: ...

    def __eq__(self, other) -> bool:
        # catch cache lookup
        if isinstance(other, _ConfigFlagBase):
            return id(other) == id(self)

        return self.get() == other


class ConfigFlag(_ConfigFlagBase[bool]):
    def __init__(self, name: str, default: bool = False, descr: str = "") -> None:
        super().__init__(name, default, descr)

    def _convert(self, raw_val: str) -> bool:
        matches = [
            (True, ["1", "true", "yes", "y"]),
            (False, ["0", "false", "no", "n"]),
        ]
        val = raw_val.lower()

        return find(matches, lambda x: val in x[1])[0]

    def __bool__(self):
        return self.get()


class ConfigFlagEnum[E: StrEnum](_ConfigFlagBase[E]):
    def __init__(self, enum: type[E], name: str, default: E, descr: str = "") -> None:
        self.enum = enum
        super().__init__(name, default, descr)

    def _convert(self, raw_val: str) -> E:
        return self.enum[raw_val.upper()]


class ConfigFlagString(_ConfigFlagBase[str]):
    def __init__(self, name: str, default: str = "", descr: str = "") -> None:
        super().__init__(name, default, descr)

    def _convert(self, raw_val: str) -> str:
        return raw_val


class ConfigFlagInt(_ConfigFlagBase[int]):
    def __init__(self, name: str, default: int = 0, descr: str = "") -> None:
        super().__init__(name, default, descr)

    def _convert(self, raw_val: str) -> int:
        if raw_val.startswith("0x"):
            return int(raw_val, 16)
        return int(float(raw_val))

    def __int__(self) -> int:
        return self.get()


class ConfigFlagFloat(_ConfigFlagBase[float]):
    def __init__(self, name: str, default: float = 0.0, descr: str = "") -> None:
        super().__init__(name, default, descr)

    def _convert(self, raw_val: str) -> float:
        return float(raw_val)

    def __float__(self) -> float:
        return self.get()


def zip_dicts_by_key(*dicts):
    keys = {k for d in dicts for k in d}
    return {k: tuple(d.get(k) for d in dicts) for k in keys}


def factory[T, **P](con: Callable[P, T]) -> Callable[P, Callable[[], T]]:
    def _(*args: P.args, **kwargs: P.kwargs) -> Callable[[], T]:
        def __() -> T:
            return con(*args, **kwargs)

        return __

    return _


class PostInitCaller(type):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post_init__(*args, **kwargs)
        return obj


def post_init_decorator(cls):
    """
    Class decorator that calls __post_init__ after the last (of derived classes)
    __init__ has been called.
    Attention: Needs to be called on cls in __init_subclass__ of decorated class.
    """
    post_init_base = getattr(cls, "__post_init_decorator", None)
    # already decorated
    if post_init_base is cls:
        return

    original_init = cls.__init__

    # inherited constructor
    if post_init_base and post_init_base.__init__ == cls.__init__:
        original_init = post_init_base.__original_init__

    def new_init(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        if hasattr(self, "__post_init__") and type(self) is cls:
            self.__post_init__(*args, **kwargs)

    cls.__init__ = new_init
    cls.__original_init__ = original_init
    cls.__post_init_decorator = cls
    return cls


class Tree[T](dict[T, "Tree[T]"]):
    def iter_by_depth(self) -> Iterable[Sequence[T]]:
        yield list(self.keys())

        for level in zip_exhaust(*[v.iter_by_depth() for v in self.values()]):
            # merge lists of parallel subtrees
            yield [n for subtree in level for n in subtree]

    def pretty(self) -> str:
        # TODO this is def broken for actual trees

        out = ""
        next_levels = [self]
        while next_levels:
            for next_level in next_levels:
                out += " | ".join(f"{p!r}" for p in next_level.keys())
            next_levels = [
                children
                for next_level in next_levels
                for _, children in next_level.items()
            ]
            if any(next_levels):
                out += indent("\n↓\n", " " * 12)

        return out

    def copy(self) -> "Tree[T]":
        return Tree({k: v.copy() for k, v in self.items()})

    def flat(self) -> set[T]:
        return {n for level in self.iter_by_depth() for n in level}

    def leaves(self) -> Generator[T, None, None]:
        for child, child_tree in self.items():
            if not child_tree:
                yield child
            else:
                yield from child_tree.leaves()

    def get_subtree(self, node: T) -> "Tree[T]":
        """
        If not acyclic, will return the highest subtree containing the node.
        """
        trees = [self]
        while trees:
            tree = find_or(
                trees,
                lambda t: node in not_none(t),
                default=None,
                default_multi=lambda x: x[0],
            )
            if tree is not None:
                return tree[node]

            trees = [child for tree in trees for child in tree.values()]

        raise KeyErrorNotFound(f"Node {node} not found in tree")


# zip iterators, but if one iterators stops producing, the rest continue
def zip_exhaust(*args):
    while True:
        out = [next(a, None) for a in args]
        out = [a for a in out if a]
        if not out:
            return

        yield out


def join_if_non_empty(sep: str, *args):
    return sep.join(s for arg in args if (s := str(arg)))


def dataclass_as_kwargs(obj: Any) -> dict[str, Any]:
    """
    Unlike dataclasses.asdict because it doesn't convert children dataclasses to dicts.
    This is useful when reconstructing a dataclass from a dict.
    """
    return {f.name: getattr(obj, f.name) for f in fields(obj)}


class RecursionGuard:
    def __init__(self, limit: int = 10000):
        self.limit = limit

    # TODO remove this workaround when we have lazy mifs
    def __enter__(self):
        self.recursion_depth = sys.getrecursionlimit()
        sys.setrecursionlimit(self.limit)

    def __exit__(self, exc_type, exc_value, traceback):
        sys.setrecursionlimit(self.recursion_depth)


def in_debug_session() -> bool:
    """
    Check if a debugger is connected.
    """
    # short-cut so we don't end up with a bunch of useless warnings
    # when just checking for debugpy in the import statement
    if "debugpy" not in sys.modules:
        return False

    try:
        import debugpy

        return debugpy.is_client_connected()

    except (ImportError, ModuleNotFoundError):
        pass

    return False


class FuncSet[T, H: Hashable = int](collections.abc.MutableSet[T]):
    """
    A set by pre-processing the objects with the hasher function.
    """

    def __init__(self, data: Iterable[T] = tuple(), hasher: Callable[[T], H] = id):
        self._hasher = hasher
        self._deref: defaultdict[H, list[T]] = defaultdict(list)

        for item in data:
            self._deref[self._hasher(item)].append(item)

    def add(self, value: T):
        if value not in self._deref[self._hasher(value)]:
            self._deref[self._hasher(value)].append(value)

    def discard(self, value: T):
        hashed = self._hasher(value)
        if hashed in self._deref and value in self._deref[hashed]:
            self._deref[hashed].remove(value)
            if not self._deref[hashed]:
                del self._deref[hashed]

    def __contains__(self, value: object):
        try:
            # Allow something of broader type than typically allowed
            # but ultimately behave the same
            hash_value = self._hasher(value)  # type: ignore
        except Exception:
            return False
        return value in self._deref[hash_value]

    def __iter__(self) -> Iterator[T]:
        yield from chain.from_iterable(self._deref.values())

    def __len__(self) -> int:
        return sum(len(v) for v in self._deref.values())

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"({repr(list(self))}, hasher={repr(self._hasher)})"
        )


class FuncDict[T, U, H: Hashable = int](collections.abc.MutableMapping[T, U]):
    """
    A dict by pre-processing the objects with the hasher function.
    """

    def __init__(
        self,
        data: Iterable[tuple[T, U]] = tuple(),
        hasher: Callable[[T], H] = id,
    ):
        self._hasher = hasher
        self._keys: defaultdict[H, list[T]] = defaultdict(list)
        self._values: defaultdict[H, list[U]] = defaultdict(list)

        for key, value in data:
            hashed_key = self._hasher(key)
            self._keys[hashed_key].append(key)
            self._values[hashed_key].append(value)

    def __contains__(self, item: object):
        try:
            hashed = self._hasher(item)  # type: ignore
        except TypeError:
            return False
        return item in self._keys[hashed]

    def keys(self) -> Iterator[T]:
        yield from chain.from_iterable(self._keys.values())

    def values(self) -> Iterator[U]:
        yield from chain.from_iterable(self._values.values())

    def __iter__(self) -> Iterator[T]:
        yield from self.keys()

    def __len__(self) -> int:
        return sum(len(v) for v in self._values.values())

    def __getitem__(self, key: T) -> U:
        hashed = self._hasher(key)
        for test_key, value in zip(self._keys[hashed], self._values[hashed]):
            if test_key == key:
                return value
        raise KeyError(key)

    def __setitem__(self, key: T, value: U):
        hashed_key = self._hasher(key)
        try:
            idx = self._keys[hashed_key].index(key)
        except ValueError:
            self._keys[hashed_key].append(key)
            self._values[hashed_key].append(value)
        else:
            self._values[hashed_key][idx] = value
            self._keys[hashed_key][idx] = key

    def __delitem__(self, key: T):
        hashed_key = self._hasher(key)
        try:
            idx = self._keys[hashed_key].index(key)
        except ValueError:
            raise KeyError(key)
        else:
            del self._values[hashed_key][idx]
            del self._keys[hashed_key][idx]

    def items(self) -> Iterable[tuple[T, U]]:
        """Iter key-value pairs as items, just like a dict."""
        yield from zip(self.keys(), self.values())

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"({repr(list(self.items()))}, hasher={repr(self._hasher)})"
        )

    def backwards_lookup(self, item: U) -> T:
        """Find the first value that maps to item, and return its key."""
        for key, value in self.items():
            if value == item:
                return key
        raise KeyError(item)

    def setdefault(self, key: T, default: U) -> U:
        """Set default if key is not in the dict, and return the value."""
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return default

    def backwards(self) -> "FuncDict[U, T, H]":
        return FuncDict(((v, k) for k, v in self.items()), hasher=self._hasher)


def dict_map_values(d: dict, function: Callable[[Any], Any]) -> dict:
    """recursively map all values in a dict"""

    result = {}
    for key, value in d.items():
        if isinstance(value, dict):
            result[key] = dict_map_values(value, function)
        elif isinstance(value, list):
            result[key] = [dict_map_values(v, function) for v in value]
        else:
            result[key] = function(value)
    return result


def merge_dicts(*dicts: dict) -> dict:
    """merge a list of dicts into a single dict,
    if same key is present and value is list, lists are merged
    if same key is dict, dicts are merged recursively
    """
    result = {}
    for d in dicts:
        for k, v in d.items():
            if k in result:
                if isinstance(v, list):
                    assert isinstance(
                        result[k], list
                    ), f"Trying to merge list into key '{k}' of type {type(result[k])}"
                    result[k] += v
                elif isinstance(v, dict):
                    assert isinstance(result[k], dict)
                    result[k] = merge_dicts(result[k], v)
                else:
                    result[k] = v
            else:
                result[k] = v
    return result


def abstract[T: type](cls: T) -> T:
    """
    Mark a class as abstract.
    """

    old_new = cls.__new__

    def _new(cls_, *args, **kwargs):
        if cls_ is cls:
            raise TypeError(f"{cls.__name__} is abstract and cannot be instantiated")
        return old_new(cls_, *args, **kwargs)

    cls.__new__ = _new
    cls.__is_abstract__ = cls
    return cls


def typename(x: object | type) -> str:
    if not isinstance(x, type):
        x = type(x)
    return x.__name__


def dict_value_visitor(d: dict, visitor: Callable[[Any, Any], Any]):
    for k, v in list(d.items()):
        if isinstance(v, dict):
            dict_value_visitor(v, visitor)
        else:
            d[k] = visitor(k, v)


def invert_dict[T, U](d: dict[T, U]) -> dict[U, list[T]]:
    return groupby(d.keys(), key=lambda k: d[k])


class DefaultFactoryDict[T, U](dict[T, U]):
    def __init__(self, factory: Callable[[T], U], *args, **kwargs):
        self.factory = factory
        super().__init__(*args, **kwargs)

    def __missing__(self, key: T) -> U:
        res = self.factory(key)
        self[key] = res
        return res


class EquivalenceClasses[T: Hashable]:
    def __init__(self, base: Iterable[T] | None = None):
        self.classes: dict[T, set[T]] = DefaultFactoryDict(lambda k: {k})
        for elem in base or []:
            self.classes[elem]

    def add_eq(self, *values: T):
        if len(values) < 2:
            return
        val1 = values[0]
        for val in values[1:]:
            self.classes[val1].update(self.classes[val])
            for v in self.classes[val]:
                self.classes[v] = self.classes[val1]

    def is_eq(self, a: T, b: T) -> bool:
        return self.classes[a] is self.classes[b]

    def get(self, only_multi: bool = False) -> list[set[T]]:
        sets = {id(s): s for s in self.classes.values()}
        if only_multi:
            sets = {k: v for k, v in sets.items() if len(v) > 1}
        return list(sets.values())


def common_prefix_to_tree(iterable: list[str]) -> Iterable[str]:
    """
    Turns:

    <760>|RP2040.adc[0]|ADC.reference|ElectricPower.max_current|Parameter
    <760>|RP2040.adc[0]|ADC.reference|ElectricPower.voltage|Parameter
    <760>|RP2040.adc[1]|ADC.reference|ElectricPower.max_current|Parameter
    <760>|RP2040.adc[1]|ADC.reference|ElectricPower.voltage|Parameter

    Into:

    <760>|RP2040.adc[0]|ADC.reference|ElectricPower.max_current|Parameter
    -----------------------------------------------.voltage|Parameter
    -----------------1]|ADC.reference|ElectricPower.max_current|Parameter
    -----------------------------------------------.voltage|Parameter

    Notes:
        Recommended to sort the iterable first.
    """
    yield iterable[0]

    for s1, s2 in pairwise(iterable):
        prefix = commonprefix([s1, s2])
        prefix_length = len(prefix)
        yield "-" * prefix_length + s2[prefix_length:]


def ind[T: str | list[str]](lines: T) -> T:
    prefix = "    "
    if isinstance(lines, str):
        return indent(lines, prefix=prefix)
    if isinstance(lines, list):
        return [f"{prefix}{line}" for line in lines]  # type: ignore


def run_live(
    *args,
    stdout: Callable[[str], Any] = logger.debug,
    stderr: Callable[[str], Any] = logger.error,
    check: bool = True,
    **kwargs,
) -> tuple[str, str, subprocess.Popen]:
    """Runs a process and logs the output live."""

    process = subprocess.Popen(
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,  # Line buffered
        **kwargs,
    )

    # Set up file descriptors to monitor
    reads = [process.stdout, process.stderr]
    stdout_lines = []
    stderr_lines = []
    while reads and process.poll() is None:
        # Wait for output on either stream
        readable, _, _ = select.select(reads, [], [])

        for stream in readable:
            line = stream.readline()
            if not line:  # EOF
                reads.remove(stream)
                continue

            if stream == process.stdout:
                stdout_lines.append(line)
                if stdout:
                    stdout(line.rstrip())
            elif stream == process.stderr:
                stderr_lines.append(line)
                if stderr:
                    stderr(line.rstrip())

    # Ensure the process has finished
    process.wait()

    # Get return code and check for errors
    if process.returncode != 0 and check:
        raise subprocess.CalledProcessError(
            process.returncode, args[0], "".join(stdout_lines), "".join(stderr_lines)
        )

    return "\n".join(stdout_lines), "\n".join(stderr_lines), process


@contextmanager
def global_lock(lock_file_path: Path, timeout_s: float | None = None):
    # TODO consider using filelock instead

    lock_file_path.parent.mkdir(parents=True, exist_ok=True)

    start_time = time.time()
    while try_or(
        lambda: bool(lock_file_path.touch(exist_ok=False)),
        default=True,
        catch=FileExistsError,
    ):
        # check if pid still alive
        try:
            pid = int(lock_file_path.read_text())
        except ValueError:
            lock_file_path.unlink(missing_ok=True)
            continue
        assert pid != os.getpid()
        if not psutil.pid_exists(pid):
            lock_file_path.unlink(missing_ok=True)
            continue
        if timeout_s and time.time() - start_time > timeout_s:
            raise TimeoutError()
        time.sleep(0.1)

    # write our pid to the lock file
    lock_file_path.write_text(str(os.getpid()))
    try:
        yield
    finally:
        lock_file_path.unlink(missing_ok=True)


def consume(iter: Iterable, n: int) -> list:
    assert n >= 0
    out = list(itertools.islice(iter, n))
    return out if len(out) == n else []


def closest_base_class(cls: type, base_classes: list[type]) -> type:
    """
    Find the most specific (closest) base class from a list of potential base classes.

    Args:
        cls: The class to find the closest base class for
        base_classes: List of potential base classes to check

    Returns:
        The most specific base class from the list that cls inherits from

    Raises:
        ValueError: If cls doesn't inherit from any of the base classes
    """
    # Get all base classes in method resolution order (most specific first)
    mro = cls.__mro__

    # Find the first (most specific) base class that appears in the provided list
    sort = sorted(base_classes, key=lambda x: mro.index(x))
    return sort[0]


def operator_type_check[**P, T](method: Callable[P, T]) -> Callable[P, T]:
    @wraps(method)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        hints = get_type_hints(
            method, include_extras=True
        )  # This resolves string annotations
        sig = inspect.signature(method)
        param_hints = {name: hint for name, hint in hints.items() if name != "return"}

        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        for name, value in bound_args.arguments.items():
            if name not in param_hints:
                continue
            expected_type = param_hints[name]
            # Handle Union, Optional, etc.
            if hasattr(expected_type, "__origin__"):
                expected_type = expected_type.__origin__
            if not isinstance(value, expected_type):
                return NotImplemented

        return method(*args, **kwargs)

    return wrapper


@overload
def partition[Y, T](
    pred: Callable[[T], TypeGuard[Y]], iterable: Iterable[T]
) -> tuple[Iterable[T], Iterable[Y]]: ...


@overload
def partition[T](
    pred: Callable[[T], bool], iterable: Iterable[T]
) -> tuple[Iterable[T], Iterable[T]]: ...


def partition(pred, iterable):  # type: ignore
    from more_itertools import partition as p

    return p(pred, iterable)


@overload
def partition_as_list[Y, T](
    pred: Callable[[T], TypeGuard[Y]], iterable: Iterable[T]
) -> tuple[list[T], list[Y]]: ...


@overload
def partition_as_list[T](
    pred: Callable[[T], bool], iterable: Iterable[T]
) -> tuple[list[T], list[T]]: ...


def partition_as_list(pred, iterable):  # type: ignore
    false_list, true_list = partition(pred, iterable)
    return list(false_list), list(true_list)


def times_out(seconds: float):
    # if running in debugger, don't timeout
    if hasattr(sys, "gettrace") and sys.gettrace():
        return lambda func: func

    def decorator[**P, T](func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            import signal

            def timeout_handler(signum, frame):
                raise TimeoutError(
                    f"Function {func.__name__} exceeded time limit of {seconds}s"
                )

            # Set up the signal handler
            old_handler = signal.signal(signal.SIGALRM, timeout_handler)
            # Set alarm to trigger after specified seconds
            signal.setitimer(signal.ITIMER_REAL, seconds)

            try:
                return func(*args, **kwargs)
            finally:
                # Cancel the alarm and restore the old signal handler
                signal.setitimer(signal.ITIMER_REAL, 0)
                signal.signal(signal.SIGALRM, old_handler)

        return wrapper

    return decorator


def hash_string(string: str) -> str:
    """Spits out a uuid in hex from a string"""
    return str(
        uuid.UUID(
            bytes=hashlib.blake2b(string.encode("utf-8"), digest_size=16).digest()
        )
    )


def get_module_from_path(
    file_path: os.PathLike, attr: str | None = None, allow_ambiguous: bool = False
) -> ModuleType | None:
    """
    Return a module based on a file path if already imported, or return None.

    If allow_ambiguous is True, and there are multiple modules with the same file path,
    return the first one.
    """
    sanitized_file_path = Path(file_path).expanduser().resolve().absolute()

    def _needle(m: ModuleType) -> bool:
        try:
            file = Path(getattr(m, "__file__"))
        except Exception:
            return False
        return sanitized_file_path.samefile(file)

    try:
        module = find(sys.modules.values(), _needle)
    except KeyErrorNotFound:
        return None
    except KeyErrorAmbiguous as e:
        if allow_ambiguous:
            module = e.duplicates[0]
        else:
            raise

    if attr is None:
        return module

    return getattr(module, attr, None)


def import_from_path(
    file_path: os.PathLike, attr: str | None = None
) -> ModuleType | Type:
    """
    Import a module from a file path.

    If the module is already imported, return the existing module.
    Otherwise, import the module and return the new module.

    Raises FileNotFoundError if the file does not exist.
    Raises AttributeError if the attr is not found in the module.
    """
    # custom unique name to avoid collisions
    # we use this hasher to generate something terse and unique
    module_name = hash_string(str(Path(file_path).expanduser().resolve().absolute()))

    if module_name in sys.modules:
        module = sys.modules[module_name]
    else:
        # setting to a sequence (and not None) indicates that the module is a package,
        # which lets us use relative imports for submodules
        submodule_search_locations = []

        spec = importlib.util.spec_from_file_location(
            module_name,
            file_path,
            submodule_search_locations=submodule_search_locations,
        )
        if spec is None:
            raise ImportError(path=str(file_path))

        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module

        assert spec.loader is not None

        spec.loader.exec_module(module)

    if attr is None:
        return module
    else:
        return getattr(module, attr)


def has_attr_or_property(obj: object, attr: str) -> bool:
    """Check if an object has an attribute or property by the name `attr`."""
    return hasattr(obj, attr) or (
        hasattr(type(obj), attr) and isinstance(getattr(type(obj), attr), property)
    )


def write_only_property(func: Callable):
    def raise_write_only(*args, **kwargs):
        raise AttributeError(f"{func.__name__} is write-only")

    return property(
        fget=raise_write_only,
        fset=func,
    )


def has_instance_settable_attr(obj: object, attr: str) -> bool:
    """
    Check if an object has an instance attribute that is settable.
    """
    # If we have a property, it's going to tell us all we need to know
    if hasattr(type(obj), attr) and isinstance(getattr(type(obj), attr), property):
        # If the property is settable, use it to set the value
        if getattr(type(obj), attr).fset is not None:
            return True
        # If not, it's not settable, end here
        return False

    # If there's an instance only attribute, we can set it
    if hasattr(obj, attr) and not hasattr(type(obj), attr):
        return True

    # If there's an instance attribute, that's unique compared to the class
    # attribute, we can set it. We don't need to check for a property here
    # because we already checked for that above.
    if (
        hasattr(obj, attr)
        and hasattr(type(obj), attr)
        and getattr(obj, attr) is not getattr(type(obj), attr)
    ):
        return True

    return False


AUTO_RECOMPILE = ConfigFlag(
    "AUTO_RECOMPILE",
    default=False,
    descr="Automatically recompile source files if they have changed",
)


# Check if installed as editable
def is_editable_install():
    distro = Distribution.from_name("atopile")

    if dist_info := distro.read_text("direct_url.json"):
        return json.loads(dist_info).get("dir_info", {}).get("editable", False)

    return False


class SerializableEnum[E: Enum](Serializable):
    class Value[E_: Enum](Serializable):
        def __init__(self, enum: "SerializableEnum", value: E_):
            self._value = value
            self._enum = enum

        @property
        def name(self) -> str:
            return self._value.name

        @property
        def value(self) -> E:
            return self._value.value

        @override
        def serialize(self) -> dict:
            return {"name": self._value.name}

        def __eq__(self, other: "SerializableEnum.Value[E]") -> bool:
            if not other._enum == self._enum:
                return False
            return (
                other._value.name == self._value.name
                and other._value.value == self._value.value
            )

        def __hash__(self) -> int:
            return hash(self._value)

        def __repr__(self) -> str:
            return f"{self._enum.enum.__name__}.{self._value.name}"

    def __init__(self, enum: type[Enum]):
        self.enum = enum

        class _Value(SerializableEnum.Value[E]):
            @override
            @staticmethod
            def deserialize(data: dict) -> "_Value":
                return _Value(self, self.enum[data["name"]])

        self.value_cls = _Value

    def serialize(self) -> dict:
        enum = self.enum
        # check enum values to all be ints or str
        if not all(isinstance(e.value, (int, str, float)) for e in enum):
            raise ValueError(f"Can't serialize {enum}: has non-primitive values")

        enum_cls_serialized = {
            "name": enum.__name__,
            "values": {e.name: e.value for e in enum},
        }

        return enum_cls_serialized

    @staticmethod
    def deserialize(data: dict) -> "SerializableEnum":
        enum_cls = Enum(data["name"], data["values"])
        return SerializableEnum(cast(type[Enum], enum_cls))

    def make_value(
        self, value: "E | SerializableEnum.Value[E]"
    ) -> "SerializableEnum.Value[E]":
        if isinstance(value, SerializableEnum.Value):
            return value
        return self.value_cls(self, value)

    def deserialize_value(self, data: dict) -> "SerializableEnum.Value[E]":
        return self.value_cls.deserialize(data)

    def __eq__(self, other: "SerializableEnum") -> bool:
        return self.enum.__name__ == other.enum.__name__ and {
            e.name: e.value for e in self.enum
        } == {e.name: e.value for e in other.enum}


def indented_container(
    obj: Iterable | dict,
    indent_level: int = 1,
    recursive: bool = False,
    use_repr: bool = True,
) -> str:
    kvs = obj.items() if isinstance(obj, dict) else list(enumerate(obj))
    _indent_prefix = "  "
    _indent = _indent_prefix * indent_level
    ind = "\n" + _indent

    def format_v(v: Any) -> str:
        if not use_repr and isinstance(v, str):
            return indent(v, prefix=_indent)
        if not recursive or not isinstance(v, Iterable) or isinstance(v, str):
            return repr(v) if use_repr else str(v)
        return indented_container(v, indent_level=indent_level + 1, recursive=recursive)

    inside = ind.join(f"{k}: {format_v(v)}" for k, v in kvs)
    if len(kvs):
        inside = f"{ind}{inside}\n"

    return f"{{{inside}{_indent_prefix * (indent_level - 1)}}}"


def md_list(
    obj: Iterable | dict, indent_level: int = 0, recursive: bool = False
) -> str:
    """
    Convert an iterable or dictionary into a nested markdown list.
    """
    indent = f"{'  ' * indent_level}"

    if isinstance(obj, dict):
        kvs = obj.items()
    elif isinstance(obj, str):
        return f"{indent}- {obj}"
    else:
        try:
            kvs = list(enumerate(obj))
        except TypeError:
            return f"{indent}- {str(obj)}"

    if not kvs:
        return f"{indent}- *(empty)*"

    lines = deque()
    for k, v in kvs:
        key_str = f" **{k}**:" if isinstance(obj, dict) else ""

        if recursive and isinstance(v, Iterable) and not isinstance(v, str):
            if isinstance(obj, dict):
                lines.append(f"{indent}-{key_str}")
            nested = md_list(v, indent_level + 1, recursive)
            lines.append(nested)
        else:
            value_str = str(v)
            lines.append(f"{indent}-{key_str} {value_str}")

    return "\n".join(lines)


def robustly_rm_dir(path: os.PathLike) -> None:
    """Remove a directory and all its contents."""

    path = Path(path)

    def remove_readonly(func, path, excinfo):
        os.chmod(path, stat.S_IWRITE)
        func(path)

    shutil.rmtree(path, onexc=remove_readonly)


def yield_missing(existing: Container, candidates: Iterable | None = None):
    if candidates is None:
        candidates = range(10000)  # Prevent counting to infinity by default
    for c in candidates:
        if c not in existing:
            yield c


def try_relative_to(
    target: os.PathLike, root: os.PathLike | None = None, walk_up: bool = False
) -> Path:
    target = Path(target)
    root = Path(root) if root is not None else Path.cwd()
    try:
        return target.relative_to(root, walk_up=walk_up)
    except FileNotFoundError:
        return target


def repo_root() -> Path:
    repo_root = Path(__file__)
    while not (repo_root / ".git").exists():
        if parent := repo_root.parent:
            repo_root = parent
        else:
            raise FileNotFoundError("Could not find repo root")
    else:
        return repo_root


def is_numeric_str(s: str) -> bool:
    """
    Check if a string is a numeric string.
    """
    return s.replace(".", "").strip().isnumeric()


def remove_venv_from_env(base_env: dict[str, str] | None = None):
    """
    Clean and return environment from venv, so subprocess can launch with system env.
    """

    env = base_env.copy() if base_env is not None else os.environ.copy()

    # Does not work, shell variables (not exported)
    # # Restore original PATH if saved
    # if "_OLD_VIRTUAL_PATH" in env:
    #     env["PATH"] = env.pop("_OLD_VIRTUAL_PATH")

    # # Restore original PYTHONHOME if saved
    # if "_OLD_VIRTUAL_PYTHONHOME" in env:
    #     env["PYTHONHOME"] = env.pop("_OLD_VIRTUAL_PYTHONHOME")

    # # Restore original shell prompt if saved (if applicable)
    # if "_OLD_VIRTUAL_PS1" in env:
    #     env["PS1"] = env.pop("_OLD_VIRTUAL_PS1")

    # Remove virtual environment specific variables
    venv = env.pop("VIRTUAL_ENV", None)
    if venv is not None:
        # Remove venv from PATH
        path = env["PATH"].split(":")
        path = [p for p in path if not p.startswith(venv)]
        env["PATH"] = ":".join(path)

    venv_prompt = env.pop("VIRTUAL_ENV_PROMPT", None)
    if venv_prompt is not None:
        # Remove venv from prompt
        prompt = env["PS1"]
        prompt = prompt.replace(venv_prompt, "")
        env["PS1"] = prompt

    env.pop("PYTHONHOME", None)

    return env



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class PoweredLED(Module):
    power: F.ElectricPower
    current_limiting_resistor: F.Resistor
    led: F.LED

    def __init__(self, low_side_resistor: bool = True):
        super().__init__()
        self._low_side_resistor = low_side_resistor

    def __preinit__(self):
        if self._low_side_resistor:
            self.power.hv.connect_via(
                [self.led, self.current_limiting_resistor], self.power.lv
            )
        else:
            self.power.hv.connect_via(
                [self.current_limiting_resistor, self.led], self.power.lv
            )

        # I = (V - V_led) / R, rearranged to assist the solver to find a solution
        self.current_limiting_resistor.resistance.alias_is(
            (self.power.voltage - self.led.forward_voltage) / self.led.current
        )

        # help solver
        self.led.forward_voltage.constrain_le(self.power.voltage)
        self.led.forward_voltage.alias_is(
            self.power.voltage
            - self.led.current * self.current_limiting_resistor.resistance
        )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.power.hv, self.power.lv)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(self.power)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.exporters.pcb.layout.heuristic_decoupling import (
    LayoutHeuristicElectricalClosenessDecouplingCaps,
)
from faebryk.exporters.pcb.layout.heuristic_pulls import (
    LayoutHeuristicElectricalClosenessPullResistors,
)
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class RaspberryPiPico(Module):
    """
    Raspberry Pi Pico clone.
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    base: F.RaspberryPiPicoBase_ReferenceDesign
    header = L.list_f_field(2, F.Header)(horizonal_pin_count=20, vertical_pin_count=1)
    usb_connector: F.USB_Type_C_Receptacle_16_pin
    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://datasheets.raspberrypi.com/pico/pico-datasheet.pdf"
    )

    @L.rt_field
    def designator_prefix(self):
        return F.has_designator_prefix(F.has_designator_prefix.Prefix.MOD)

    @L.rt_field
    def pcb_layout(self):
        from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
        from faebryk.exporters.pcb.layout.extrude import LayoutExtrude
        from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy

        Point = F.has_pcb_position.Point
        L = F.has_pcb_position.layer_type
        LVL = LayoutTypeHierarchy.Level

        layout = F.has_pcb_layout_defined(
            layout=LayoutTypeHierarchy(
                layouts=[
                    LVL(
                        mod_type=F.RaspberryPiPicoBase_ReferenceDesign,
                        layout=LayoutAbsolute(
                            Point((0, 0, 0, L.NONE)),
                        ),
                    ),
                    LVL(
                        mod_type=F.Header,
                        layout=LayoutExtrude(
                            base=Point((0, 0, L.NONE)), vector=(17.78, 0)
                        ),
                    ),
                    LVL(
                        mod_type=F.USB_Type_C_Receptacle_16_pin,
                        layout=LayoutAbsolute(
                            Point((17.78 / 2, 0, L.NONE)),
                        ),
                    ),
                ]
            )
        )

        LayoutHeuristicElectricalClosenessDecouplingCaps.add_to_all_suitable_modules(
            self
        )
        LayoutHeuristicElectricalClosenessPullResistors.add_to_all_suitable_modules(
            self
        )

        return layout

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        power_3v3 = self.base.ldo.power_out
        gnd = power_3v3.lv

        gpio_count = 0
        for i, pin in enumerate(self.header[0].contact):
            if i in [2, 7, 12, 17]:
                pin.connect(gnd)
            else:
                pin.connect(self.base.rp2040.gpio[gpio_count].line)
                self.base.rp2040.pinmux.enable(self.base.rp2040.gpio[gpio_count])
                gpio_count += 1
        for i, pin in enumerate(self.header[1].contact):
            if i in [2, 7, 12, 17]:
                pin.connect(gnd)
            elif i == 9:
                pin.connect(self.base.rp2040.run.line)
            elif i == 14:
                ...  # TODO: ADC_VREF is not implemented
            elif i == 15:
                pin.connect(power_3v3.hv)
            elif i == 16:
                pin.connect(self.base.ldo.enable.get_enable_signal())
            elif i == 18:
                pin.connect(self.base.ldo.power_in.hv)
            elif i == 19:
                pin.connect(self.base.usb.usb_if.buspower.hv)
            else:
                if gpio_count == 23:
                    gpio_count += 3  # skip 23, 24, 25
                pin.connect(self.base.rp2040.gpio[gpio_count].line)
                self.base.rp2040.pinmux.enable(self.base.rp2040.gpio[gpio_count])
                gpio_count += 1

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        for header in self.header:
            header.pin_pitch.constrain_subset(2.54 * P.mm)
            header.mating_pin_length.constrain_subset(
                L.Range.from_center_rel(6 * P.mm, 0.1)
            )
            header.pad_type.constrain_subset(F.Header.PadType.THROUGH_HOLE)
            header.pin_type.constrain_subset(F.Header.PinType.MALE)
            header.angle.constrain_subset(F.Header.Angle.STRAIGHT)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import TYPE_CHECKING, Callable

from faebryk.libs.library import L

if TYPE_CHECKING:
    from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer

logger = logging.getLogger(__name__)


class has_layout_transform(L.Module.TraitT.decless()):
    """
    Docstring describing your module
    """

    def __init__(self, transform: Callable[["PCB_Transformer"], None]):
        super().__init__()
        self.transform = transform



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class USB_RS485(Module):
    usb_uart: F.CH340x
    uart_rs485: F.UART_RS485
    termination: F.Resistor
    polarization = L.list_field(2, F.Resistor)
    usb: F.USB2_0
    rs485: F.RS485HalfDuplex

    def __preinit__(self):
        self.usb.connect(self.usb_uart.usb)
        self.usb_uart.uart.base_uart.connect(self.uart_rs485.uart)
        self.rs485.connect(self.uart_rs485.rs485)

        self.usb_uart.tnow.connect(self.uart_rs485.read_enable)
        self.usb_uart.tnow.connect(self.uart_rs485.write_enable)

        self.usb_uart.usb.usb_if.buspower.connect(self.uart_rs485.power)
        self.usb.usb_if.buspower.connect(self.usb_uart.usb.usb_if.buspower)

        # connect termination resistor between RS485 A and B
        self.uart_rs485.rs485.diff_pair.n.line.connect_via(
            self.termination, self.uart_rs485.rs485.diff_pair.p.line
        )

        # connect polarization resistors to RS485 A and B
        self.uart_rs485.rs485.diff_pair.p.line.connect_via(
            self.polarization[0],
            self.uart_rs485.power.hv,
        )
        self.uart_rs485.rs485.diff_pair.n.line.connect_via(
            self.polarization[1],
            self.uart_rs485.power.lv,
        )

        self.termination.resistance.constrain_subset(
            L.Range.from_center(150 * P.ohm, 1.5 * P.ohm)
        )
        self.polarization[0].resistance.constrain_subset(
            L.Range.from_center(680 * P.ohm, 6.8 * P.ohm)
        )
        self.polarization[1].resistance.constrain_subset(
            L.Range.from_center(680 * P.ohm, 6.8 * P.ohm)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging
from itertools import pairwise
from typing import (
    Iterable,
    Self,
    Sequence,
    cast,
)

from faebryk.core.cpp import (
    GraphInterfaceModuleConnection,
    Path,
)
from faebryk.core.graphinterface import GraphInterface
from faebryk.core.link import (
    Link,
    LinkDirect,
    LinkDirectConditional,
    LinkDirectConditionalFilterResult,
    LinkDirectDerived,
)
from faebryk.core.node import CNode, Node, NodeException
from faebryk.core.pathfinder import find_paths
from faebryk.core.trait import Trait
from faebryk.library.can_specialize import can_specialize
from faebryk.libs.util import ConfigFlag, cast_assert, groupby, once

logger = logging.getLogger(__name__)


IMPLIED_PATHS = ConfigFlag("IMPLIED_PATHS", default=False, descr="Use implied paths")

type Bridgable[T: "ModuleInterface"] = Node | T


class ModuleInterface(Node):
    class TraitT(Trait): ...

    specializes: GraphInterface
    specialized: GraphInterface
    connected: GraphInterfaceModuleConnection

    # TODO: move to cpp
    class _LinkDirectShallow(LinkDirectConditional):
        """
        Make link that only connects up but not down
        """

        def is_childtype_of_test_type(self, node: CNode):
            return isinstance(node, self.children_types)
            # return type(node) in self.children_types

        def check_path(self, path: Path) -> LinkDirectConditionalFilterResult:
            out = (
                LinkDirectConditionalFilterResult.FILTER_PASS
                if not self.is_childtype_of_test_type(path[0].node)
                else LinkDirectConditionalFilterResult.FILTER_FAIL_UNRECOVERABLE
            )
            return out

        def __init__(self, test_type: type["ModuleInterface"]):
            self.test_type = test_type
            # TODO this is a bit of a hack to get the children types
            #  better to do on set_connections
            self.children_types = tuple(
                type(c)
                for c in test_type().get_children(
                    direct_only=False, types=ModuleInterface, include_root=False
                )
            )
            super().__init__(
                self.check_path,
                needs_only_first_in_path=True,
            )

    @classmethod
    @once
    def LinkDirectShallow(cls):
        class _LinkDirectShallowMif(ModuleInterface._LinkDirectShallow):
            def __init__(self):
                super().__init__(test_type=cls)

        return _LinkDirectShallowMif

    def __preinit__(self) -> None: ...

    def connect(
        self: Self, *other: Self, link: type[Link] | Link | None = None
    ) -> Self:
        mismatched = [o for o in other if type(o) is not type(self)]
        if mismatched:
            if len(mismatched) == 1:
                raise NodeException(
                    self,
                    (
                        f"Cannot connect module `{self}` "
                        f"(type `{type(self).__name__}`) "
                        f"with module `{mismatched[0]}` "
                        f"(type `{type(mismatched[0]).__name__}`). "
                        "Modules must be of the same type."
                    ),
                )
            else:
                mismatches_str = ", ".join(
                    f"{m} (type {type(m).__name__})" for m in mismatched
                )
                raise NodeException(
                    self,
                    (
                        f"Cannot connect module {self} (type {type(self).__name__}) "
                        f"with modules: {mismatches_str}. "
                        "Modules must be of the same type."
                    ),
                )

        # TODO: consider returning self always
        # - con: if construcing anonymous stuff in connection no ref
        # - pro: more intuitive
        ret = other[-1] if other else self

        if link is None:
            link = LinkDirect
        if isinstance(link, type):
            link = link()

        # resolve duplicate links
        new_links = [
            o.connected
            for o in other
            if not (existing_link := self.connected.is_connected_to(o.connected))
            or existing_link != link
        ]

        self.connected.connect(new_links, link=link)

        return ret

    def connect_via(
        self,
        bridge: Bridgable[Self] | Sequence[Bridgable[Self]],
        *other: Self,
        link=None,
    ):
        from faebryk.library.can_bridge import can_bridge

        bridges = [bridge] if isinstance(bridge, Node | ModuleInterface) else bridge
        intf = self
        for sub_bridge in bridges:
            if isinstance(sub_bridge, type(self)):
                intf.connect(sub_bridge, link=link)
            else:
                t = sub_bridge.get_trait(can_bridge)
                intf.connect(t.get_in(), link=link)
                intf = t.get_out()

        intf.connect(*other, link=link)

    def connect_shallow(self, *other: Self) -> Self:
        # TODO: clone limitation, waiting for c++ LinkShallow
        if len(other) > 1:
            for o in other:
                self.connect_shallow(o)
            return self

        return self.connect(*other, link=type(self).LinkDirectShallow())

    def get_connected(self, include_self: bool = False) -> dict[Self, Path]:
        paths = find_paths(self, [])
        # TODO theoretically we could get multiple paths for the same MIF
        # practically this won't happen in the current implementation
        paths_per_mif = groupby(paths, lambda p: cast_assert(type(self), p[-1].node))

        def choose_path(_paths: list[Path]) -> Path:
            return self._path_with_least_conditionals(_paths)

        path_per_mif = {
            mif: choose_path(paths)
            for mif, paths in paths_per_mif.items()
            if mif is not self or include_self
        }
        if include_self:
            assert self in path_per_mif
        for mif, paths in paths_per_mif.items():
            self._connect_via_implied_paths(mif, paths)
        return path_per_mif

    def is_connected_to(self, other: "ModuleInterface") -> list[Path]:
        return [
            path for path in find_paths(self, [other]) if path[-1] is other.self_gif
        ]

    def specialize[T: ModuleInterface](self, special: T) -> T:
        logger.debug(f"Specializing MIF {self} with {special}")

        extra = set()
        # allow non-base specialization if explicitly allowed
        if special.has_trait(can_specialize):
            extra = set(special.get_trait(can_specialize).get_specializable_types())

        assert isinstance(special, type(self)) or any(
            issubclass(t, type(self)) for t in extra
        )

        # This is doing the heavy lifting
        self.connected.connect(special.connected)

        # Establish sibling relationship
        self.specialized.connect(special.specializes)

        return cast(T, special)

    # def get_general(self):
    #    out = self.specializes.get_parent()
    #    if out:
    #        return out[0]
    #    return None

    def __init_subclass__(cls, *, init: bool = True) -> None:
        if hasattr(cls, "_on_connect"):
            raise TypeError("Overriding _on_connect is deprecated")

        return super().__init_subclass__(init=init)

    @staticmethod
    def _path_with_least_conditionals(paths: list["Path"]) -> "Path":
        if len(paths) == 1:
            return paths[0]

        paths_links = [
            (
                path,
                [
                    e1.is_connected_to(e2)
                    for e1, e2 in pairwise(cast(Iterable[GraphInterface], path))
                ],
            )
            for path in paths
        ]
        paths_conditionals = [
            (
                path,
                [link for link in links if isinstance(link, LinkDirectConditional)],
            )
            for path, links in paths_links
        ]
        path = min(paths_conditionals, key=lambda x: len(x[1]))[0]
        return path

    def _connect_via_implied_paths(self, other: Self, paths: list["Path"]):
        if not IMPLIED_PATHS:
            return

        if self is other:
            return

        if self.connected.is_connected_to(other.connected):
            # TODO link resolution
            return

        # heuristic: choose path with fewest conditionals
        path = self._path_with_least_conditionals(paths)

        self.connect(other, link=LinkDirectDerived(path))

    @staticmethod
    def _group_into_buses[T: ModuleInterface](mifs: Iterable[T]) -> dict[T, set[T]]:
        """
        returns dict[BusRepresentative, set[MIFs in Bus]]
        """
        to_check = set(mifs)
        buses = {}
        while to_check:
            interface = to_check.pop()
            ifs = interface.get_connected(include_self=True)
            buses[interface] = ifs
            to_check.difference_update(ifs.keys())

        return buses



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties


class TXS0102DCUR(Module):
    """
    TXS0102 2-Bit Bidirectional Voltage-Level Translator for
    Open-Drain and Push-Pull Applications
    """

    class _BidirectionalLevelShifter(Module):
        # interfaces
        io_a: F.ElectricLogic
        io_b: F.ElectricLogic

        # TODO: bridge shallow

        @L.rt_field
        def can_bridge(self):
            return F.can_bridge_defined(self.io_a, self.io_b)

        # interfaces

    voltage_a_power: F.ElectricPower
    voltage_b_power: F.ElectricPower
    n_oe: F.ElectricLogic

    shifters = L.list_field(2, _BidirectionalLevelShifter)

    def __preinit__(self):
        gnd = self.voltage_a_power.lv
        gnd.connect(self.voltage_b_power.lv)

        # FIXME
        # self.voltage_a_power.decoupled.decouple()
        # self.voltage_b_power.decoupled.decouple()

        # eo is referenced to voltage_a_power (active high)
        self.n_oe.reference.connect(self.voltage_a_power)

        for shifter in self.shifters:
            side_a = shifter.io_a
            # side_a.reference.connect(self.voltage_a_power)
            side_a.add(F.has_single_electric_reference_defined(self.voltage_a_power))
            side_b = shifter.io_b
            # side_b.reference.connect(self.voltage_b_power)
            side_b.add(F.has_single_electric_reference_defined(self.voltage_b_power))

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://datasheet.lcsc.com/lcsc/1810292010_Texas-Instruments-TXS0102DCUR_C53434.pdf"
    )

    mfr = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "Texas Instruments",
            DescriptiveProperties.partno: "TXS0102DCUR",
        }
    )

    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C53434"})

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.voltage_a_power.lv: ["GND"],
                self.voltage_a_power.hv: ["VCCA"],
                self.voltage_b_power.hv: ["VCCB"],
                self.n_oe.line: ["OE"],
                self.shifters[0].io_a.line: ["A1"],
                self.shifters[0].io_b.line: ["B1"],
                self.shifters[1].io_a.line: ["A2"],
                self.shifters[1].io_b.line: ["B2"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )



"""
We collect anonymous telemetry data to help improve atopile.
To opt out, add `telemetry: false` to your project's ~/atopile/telemetry.yaml file.

What we collect:
- Hashed user id so we know how many unique users we have
- Hashed project id
- Error logs
- How long the build took
- ato version
- Git has of current commit
"""

import hashlib
import logging
import subprocess
import time
from pathlib import Path
from typing import Optional

import requests
from attrs import asdict, define, field
from ruamel.yaml import YAML

log = logging.getLogger(__name__)


@define
class TelemetryData:
    project_id: Optional[str]
    user_id: Optional[str]
    git_hash: Optional[str]
    subcommand: Optional[str]
    time: Optional[float] = None
    assertions: list = field(factory=list)
    eqns_vars: int = 0
    assertions_checked: int = 0
    ato_error: int = 0
    crash: int = 0


telemetry_data: TelemetryData | None = None
start_time: Optional[float] = None


def setup_telemetry_data(command: str):
    global telemetry_data
    _start_timer()
    telemetry_data = TelemetryData(
        project_id=get_project_id(),
        user_id=get_user_id(),
        git_hash=get_current_git_hash(),
        subcommand=command,
    )


def log_telemetry():
    try:
        # Check if telemetry is enabled
        if not load_telemetry_setting():
            log.log(0, "Telemetry is disabled. Skipping telemetry logging.")
            return
        telemetry_data.time = _end_timer()
        telemetry_dict = asdict(telemetry_data)
        log.log(0, "Logging telemetry data %s", telemetry_dict)
        requests.post(
            "https://log-telemetry-atsuhzfd5a-uc.a.run.app",
            json=telemetry_dict,
            timeout=0.1,
        ).raise_for_status()
    except requests.exceptions.Timeout:
        # We specifically ignore timeouts here because we expect
        # them to happen most of the time.
        # It's not actually a problem because we don't need to know
        # if the telemetry was logged and we don't want to slow atopile
        # down for it
        pass
    except Exception as e:
        log.log(0, "Failed to log telemetry data: %s", e)


def load_telemetry_setting() -> dict:
    atopile_home = Path.home() / ".atopile"
    atopile_yaml = atopile_home / "telemetry.yaml"

    if not atopile_yaml.exists():
        atopile_home.mkdir(parents=True, exist_ok=True)
        with atopile_yaml.open("w") as f:
            yaml = YAML()
            yaml.dump({"telemetry": True}, f)
        return True

    with atopile_yaml.open() as f:
        yaml = YAML()
        config = yaml.load(f)
        return config.get("telemetry", True)


def _start_timer():
    global start_time  # Declare start_time as global
    start_time = time.time()


def _end_timer():
    try:
        if start_time is None:
            log.log(0, "Timer was not started.")
            return -1
        end_time = time.time()
        execution_time = end_time - start_time
        log.log(0, "Execution time: %s", execution_time)
    except Exception as ex:
        log.log(0, "Failed to get execution time: %s", ex)
        return None
    return execution_time


def get_user_id() -> str:
    """Generate a unique user ID from the git email."""
    try:
        git_email = (
            subprocess.check_output(["git", "config", "user.email"])
            .decode("ascii")
            .strip()
        )
    except subprocess.CalledProcessError:
        git_email = "unknown"
    return git_email


def get_current_git_hash() -> Optional[str]:
    """Get the current git commit hash."""
    try:
        return (
            subprocess.check_output(
                ["git", "rev-parse", "HEAD"], stderr=subprocess.DEVNULL
            )
            .decode("ascii")
            .strip()
        )

    except subprocess.CalledProcessError:
        return None


def commonise_project_url(git_url: str) -> str:
    """
    Commonize the remote which could be in either of these forms:
        - https://github.com/atopile/atopile.git
        - git@github.com:atopile/atopile.git
    ... to "github.com/atopile/atopile"
    """

    if git_url.startswith("https://"):
        git_url = git_url[8:]
    elif git_url.startswith("git@"):
        git_url = git_url[4:]
        git_url = "/".join(git_url.split(":", 1))

    if git_url.endswith(".git"):
        git_url = git_url[:-4]

    return git_url


def get_project_id() -> Optional[str]:
    """Get the hashed project ID from the git URL of the project, if not available, return 'none'."""  # noqa: E501  # pre-existing
    try:
        git_url = (
            subprocess.check_output(["git", "config", "--get", "remote.origin.url"])
            .decode("ascii")
            .strip()
        )
        if not git_url:
            return None

        project_url = commonise_project_url(git_url)

        log.log(0, "Project URL: %s", project_url)

        # Hash the project ID to de-identify it
        return hashlib.sha256(project_url.encode()).hexdigest()

    except subprocess.CalledProcessError:
        return None



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import json
import logging
from pathlib import Path

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.exporters.pcb.routing.util import get_internal_nets_of_node

logger = logging.getLogger(__name__)


def _get_testpoints(app: Module) -> list[F.TestPoint]:
    return [
        testpoint
        for testpoint in app.get_children_modules(
            types=F.TestPoint,
            include_root=True,
        )
        if testpoint.has_trait(F.has_footprint)
        and testpoint.get_trait(F.has_footprint)
        .get_footprint()
        .has_trait(F.has_kicad_footprint)
    ]


def export_testpoints(
    app: Module,
    testpoints_file: Path,
):
    """
    Get all testpoints from the application and export their information to a JSON file
    """
    testpoint_data: dict[str, dict] = {}
    testpoints = _get_testpoints(app)

    for testpoint in testpoints:
        designator = testpoint.get_trait(F.has_designator).get_designator()
        full_name = testpoint.get_full_name()
        fp = testpoint.get_trait(F.has_footprint).get_footprint()
        footprint = PCB_Transformer.get_fp(fp)  # get KiCad footprint
        position = footprint.at
        layer = footprint.layer
        library_name = footprint.name

        # Get single connected net name
        nets = get_internal_nets_of_node(testpoint)
        net_name = next(
            (
                net.get_trait(F.has_overriden_name).get_name()
                for net, _ in nets.items()
                if net
            ),
            "no-net",
        )

        testpoint_data[designator] = {
            "testpoint_name": full_name,
            "net_name": net_name,
            "footprint_name": library_name,
            "position": {"x": position.x, "y": position.y, "rotation": position.r},
            "layer": layer,
        }

    with open(testpoints_file, "w") as f:
        json.dump(obj=testpoint_data, fp=f, indent=4)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class ESP32_C3(Module):
    """ESP32-C3"""

    vdd3p3_cpu: F.ElectricPower
    vdd3p3_rtc: F.ElectricPower
    vdd_spi: F.ElectricPower
    vdd3p3: F.ElectricPower
    vdda: F.ElectricPower
    lna_in: F.Electrical
    enable: F.ElectricLogic
    xtal_p: F.Electrical
    xtal_n: F.Electrical
    gpio = L.list_field(22, F.ElectricLogic)
    # TODO: map peripherals to GPIOs with pinmux
    usb: F.USB2_0
    i2c: F.I2C
    uart = L.list_field(2, F.UART_Base)
    # ... etc

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://www.espressif.com/sites/default/files/documentation/esp32-c3_datasheet_en.pdf"
    )

    def __preinit__(self):
        x = self

        # https://www.espressif.com/sites/default/files/documentation/esp32-c3_technical_reference_manual_en.pdf#uart
        for ser in x.uart:
            ser.baud.constrain_le(5 * P.mbaud)

        # connect all logic references
        # TODO: set correctly for each power domain
        # ref = F.ElectricLogic.connect_all_module_references(self)
        # self.add(F.has_single_electric_reference_defined(ref))

        # set power domain constraints to recommended operating conditions
        for power_domain in [self.vdd3p3_rtc, self.vdd3p3, self.vdda]:
            power_domain.voltage.constrain_subset(
                L.Range.from_center(3.3 * P.V, 0.3 * P.V)
            )
        self.vdd3p3_cpu.voltage.constrain_subset(
            L.Range(3.0 * P.V, 3.6 * P.V)
        )  # TODO: max 3.3V when writing eFuses
        self.vdd_spi.voltage.constrain_subset(
            L.Range.from_center(3.3 * P.V, 0.3 * P.V)
        )  # TODO: when configured as input

        # connect all grounds to eachother and power
        self.vdd3p3.lv.connect(
            self.vdd3p3_cpu.lv,
            self.vdd3p3_rtc.lv,
            self.vdda.lv,
            self.vdd_spi.lv,
        )

        # FIXME: this has to be done in ReferenceDesign or parent
        # connect decoupling caps to power domains
        # self.vdd3p3.decoupled.decouple()
        # self.vdd3p3_cpu.decoupled.decouple()
        # self.vdd3p3_rtc.decoupled.decouple()
        # self.vdda.decoupled.decouple()
        # self.vdd_spi.decoupled.decouple()

        # rc delay circuit on enable pin for startup delay
        # https://www.espressif.com/sites/default/files/documentation/esp32-c3-mini-1_datasheet_en.pdf page 24  # noqa E501
        # TODO: add lowpass filter
        # self.enable.line.connect_via(
        #    self.en_rc_capacitor, self.pwr3v3.lv
        # )
        # FIXME: this has to be done in ReferenceDesign or parent
        # self.enable.pulled.pull(up=True)  # TODO: combine with lowpass filter

    # TODO: Fix this
    #    # set mux states
    #    # UART 1
    #    self.set_mux(x.gpio[20], self.serial[1].rx)
    #    self.set_mux(x.gpio[21], self.serial[1].tx)
    #    # UART 0
    #    self.set_mux(x.gpio[0], self.serial[0].rx)
    #    self.set_mux(x.gpio[1], self.serial[0].tx)

    #    # F.I2C
    #    self.set_mux(x.gpio[4], self.i2c.scl)
    #    self.set_mux(x.gpio[5], self.i2c.sda)

    #    class _uart_esphome_config(has_esphome_config.impl()):
    #        def get_config(self_) -> dict:
    #            assert isinstance(self, ESP32_C3)
    #            obj = self_.obj
    #            assert isinstance(obj, F.UART_Base)
    #            config = {
    #                "uart": [
    #                    {
    #                        "id": obj.get_trait(is_esphome_bus).get_bus_id(),
    #                        "baud_rate": get_parameter_max(obj.baud),
    #                    }
    #                ]
    #            }

    #            try:
    #                config["uart"][0]["rx_pin"] = self.get_mux_pin(obj.rx)[1]
    #            except IndexError:
    #                ...

    #            try:
    #                config["uart"][0]["tx_pin"] = self.get_mux_pin(obj.tx)[1]
    #            except IndexError:
    #                ...

    #            # if no rx/tx pin is set, then not in use
    #            if set(config["uart"][0].keys()).isdisjoint({"rx_pin", "tx_pin"}):
    #                return {}

    #            return config

    #    class _i2c_esphome_config(has_esphome_config.impl()):
    #        def get_config(self_) -> dict:
    #            assert isinstance(self, ESP32_C3)
    #            obj = self_.obj
    #            assert isinstance(obj, F.I2C)

    #            try:
    #                sda = self.get_mux_pin(obj.sda)[1]
    #                scl = self.get_mux_pin(obj.scl)[1]
    #            except IndexError:
    #                # Not in use if pinmux is not set
    #                return {}

    #            config = {
    #                "i2c": [
    #                    {
    #                        "id": obj.get_trait(is_esphome_bus).get_bus_id(),
    #                        "frequency": int(get_parameter_max(obj.frequency)), # noqa: E501
    #                        "sda": sda,
    #                        "scl": scl,
    #                    }
    #                ]
    #            }

    #            return config

    #    for serial in self.serial:
    #        serial.add(
    #            is_esphome_bus_defined(f"uart_{self.serial.index(serial)}")
    #        )
    #        serial.add(_uart_esphome_config())

    #    for i, gpio in enumerate(self.gpio):
    #        gpio.add(is_esphome_bus_defined(f"GPIO{i}"))

    #    self.i2c.add(is_esphome_bus_defined("i2c_0"))
    #    self.i2c.add(_i2c_esphome_config())
    #    self.i2c.frequency.constrain_subset(
    #        Set(
    #            [
    #                F.I2C.define_max_frequency_capability(speed)
    #                for speed in [
    #                    F.I2C.SpeedMode.low_speed,
    #                    F.I2C.SpeedMode.standard_speed,
    #                ]
    #            ]
    #            + [
    #                L.Range(10 * P.khertz, 800 * P.khertz)
    #            ],  # TODO: should be range 200k-800k, but breaks parameter merge
    #        )
    #    )

    #    self.add(
    #        has_esphome_config_defined(
    #            {
    #                "esp32": {
    #                    "board": "Espressif ESP32-C3-DevKitM-1",
    #                    "variant": "esp32c3",
    #                    "framework": {
    #                        "type": "esp-idf",
    #                        "version": "recommended",
    #                    },
    #                },
    #            }
    #        )
    #    )

    # very simple mux that uses pinmap
    # def set_mux(self, gpio: F.ElectricLogic, target: F.ElectricLogic):
    #    """Careful not checked"""
    #    pin, _ = self.get_mux_pin(gpio)
    #    self.pinmap[pin] = target.line

    # def get_mux_pin(self, target: F.ElectricLogic) -> tuple[str, int]:
    #    """Returns pin & gpio number"""
    #    pin = [k for k, v in self.pinmap.items() if v == target.line][0]
    #    gpio = self.pinmap_default[pin]
    #    gpio_index = [
    #        i for i, g in enumerate(self.gpio) if g.line == gpio
    #    ][0]
    #    return pin, gpio_index

    def set_default_boot_mode(
        self, owner: Module, default_boot_to_spi_flash: bool = True
    ):
        # set default boot mode to "SPI Boot mode"
        # https://www.espressif.com/sites/default/files/documentation/esp32-c3_datasheet_en.pdf page 26 # noqa E501
        # TODO: make configurable
        self.gpio[8].pulled.pull(up=True, owner=owner).resistance.constrain_subset(
            L.Range.from_center_rel(10 * P.kohm, 0.1)
        )
        self.gpio[2].pulled.pull(up=True, owner=owner).resistance.constrain_subset(
            L.Range.from_center_rel(10 * P.kohm, 0.1)
        )
        # gpio[9] has an internal pull-up at boot = SPI-Boot
        if not default_boot_to_spi_flash:
            self.gpio[9].pulled.pull(up=False, owner=owner).resistance.constrain_subset(
                L.Range.from_center_rel(10 * P.kohm, 0.1)
            )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

logger = logging.getLogger(__name__)


def prettify_sexp_string(raw: str) -> str:
    out = ""
    level = 0
    in_quotes = False
    for c in raw:
        if c == '"':
            in_quotes = not in_quotes
        if in_quotes:
            ...
        elif c == "\n":
            continue
        elif c == " " and out[-1] == " ":
            continue
        elif c == "(":
            if level != 0:
                out += "\n" + " " * 4 * level
            level += 1
        elif c == ")":
            level -= 1
        out += c

    # if i > 0 no strip is a kicad bug(?) workaround
    out = "\n".join(x.rstrip() if i > 0 else x for i, x in enumerate(out.splitlines()))
    return out
